#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import os
import aiohttp
import copy
import logging
import typing
import ccxt
import ccxt.pro as ccxt_pro
import ccxt.async_support as async_ccxt

import octobot_commons.time_frame_manager as time_frame_manager
import octobot_commons.aiohttp_util as aiohttp_util
import octobot_commons.logging as commons_logging
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.exchanges.connectors.ccxt.enums as ccxt_enums
import octobot_trading.exchanges.connectors.ccxt.ccxt_clients_cache as ccxt_clients_cache
import octobot_trading.exchanges.config.proxy_config as proxy_config_import
import octobot_trading.exchanges.util.exchange_util as exchange_util


def create_client(
    exchange_class, exchange_manager, logger, options, headers,
    additional_config, should_authenticate, unauthenticated_exchange_fallback=None,
    keys_adapter=None, allow_request_counter: bool = True
):
    """
    Exchange instance creation
    :return: the created ccxt (pro, async or sync) client
    """
    is_authenticated = False
    if not exchange_manager.exchange_only:
        # avoid logging version on temporary exchange_only exchanges
        exchange_type = exchange_util.get_exchange_type(exchange_manager)
        logger.info(f"Creating {exchange_class.__name__} {exchange_type.name} "
                    f"exchange with ccxt in version {ccxt.__version__}")
    if exchange_manager.ignore_config or exchange_manager.check_config(exchange_manager.exchange_name):
        try:
            auth_token_header_prefix = None
            key, secret, password, uid, auth_token = exchange_manager.get_exchange_credentials(
                exchange_manager.exchange_name
            )
            if keys_adapter:
                key, secret, password, uid, auth_token, auth_token_header_prefix = keys_adapter(
                    key, secret, password, uid, auth_token
                )
            if not (key and secret) and not exchange_manager.is_simulated and not exchange_manager.ignore_config:
                logger.warning(f"No exchange API key set for {exchange_manager.exchange_name}. "
                               f"Enter your account details to enable real trading on this exchange.")
            if should_authenticate and not exchange_manager.is_backtesting:
                client = instantiate_exchange(
                    exchange_class,
                    _get_client_config(
                        exchange_class, options, headers, additional_config,
                        api_key=key, secret=secret, password=password, uid=uid,
                        auth_token=auth_token, auth_token_header_prefix=auth_token_header_prefix
                    ),
                    exchange_manager.exchange_name,
                    exchange_manager.proxy_config,
                    allow_request_counter=allow_request_counter
                )
                is_authenticated = True
                if exchange_manager.check_credentials:
                    client.check_required_credentials()
            else:
                client = instantiate_exchange(
                    exchange_class,
                    _get_client_config(exchange_class, options, headers, additional_config),
                    exchange_manager.exchange_name,
                    exchange_manager.proxy_config,
                    allow_request_counter=allow_request_counter
                )
        except (ccxt.AuthenticationError, Exception) as e:
            if unauthenticated_exchange_fallback is None:
                return get_unauthenticated_exchange(
                    exchange_class, options, headers, additional_config,
                    exchange_manager.exchange_name, exchange_manager.proxy_config
                ), False
            return unauthenticated_exchange_fallback(e), False
    else:
        client = get_unauthenticated_exchange(
            exchange_class, options, headers, additional_config,
            exchange_manager.exchange_name, exchange_manager.proxy_config
        )
        logger.error("configuration issue: missing login information !")
    client.logger.setLevel(logging.INFO)
    return client, is_authenticated


async def close_client(client):
    await client.close()
    client.markets = {}
    client.markets_by_id = {}
    client.ids = []
    client.last_json_response = {}
    client.last_http_response = ""
    client.last_response_headers = {}
    client.markets_loading = None
    client.currencies = {}
    client.baseCurrencies = {}
    client.quoteCurrencies = {}
    client.currencies_by_id = {}
    client.codes = []
    client.symbols = {}
    client.accounts = []
    client.accounts_by_id = {}
    client.ohlcvs = {}
    client.trades = {}
    client.orderbooks = {}


def get_unauthenticated_exchange(
    exchange_class, options, headers, additional_config, identifier: str, proxy_config: proxy_config_import.ProxyConfig
) -> async_ccxt.Exchange:
    return instantiate_exchange(
        exchange_class,
        _get_client_config(exchange_class, options, headers, additional_config),
        identifier,
        proxy_config
    )


def instantiate_exchange(
    exchange_class, config: dict, identifier: str, proxy_config: proxy_config_import.ProxyConfig,
    allow_request_counter: bool = True
) -> async_ccxt.Exchange:
    client = exchange_class(config)
    _use_proxy_if_necessary(client, proxy_config)
    if constants.ENABLE_CCXT_REQUESTS_COUNTER and allow_request_counter:
        _use_request_counter(identifier, client)
    return client


def set_sandbox_mode(exchange_connector, is_sandboxed):
    try:
        exchange_connector.client.setSandboxMode(is_sandboxed)
    except ccxt.NotSupported as e:
        default_type = exchange_connector.client.options.get('defaultType', None)
        additional_info = f" in type {default_type}" if default_type else ""
        exchange_connector.logger.warning(f"{exchange_connector.name} does not support sandboxing {additional_info}: {e}")
        # raise exception to stop this exchange and prevent dealing with a real funds exchange
        raise e
    return None


def load_markets_from_cache(client, market_filter: typing.Union[None, typing.Callable[[dict], bool]] = None):
    client.set_markets(
        market
        for market in ccxt_clients_cache.get_exchange_parsed_markets(ccxt_clients_cache.get_client_key(client))
        if market_filter is None or market_filter(market)
    )


def set_markets_cache(client):
    if client.markets:
        ccxt_clients_cache.set_exchange_parsed_markets(
            ccxt_clients_cache.get_client_key(client), copy.deepcopy(list(client.markets.values()))
        )


def get_ccxt_client_login_options(exchange_manager):
    """
    :return: ccxt client login option dict, can be overwritten to custom exchange login
    """
    if exchange_manager.is_future:
        return {'defaultType': 'future'}
    if exchange_manager.is_margin:
        return {'defaultType': 'margin'}
    return {'defaultType': 'spot'}


def get_symbols(client, active_only) -> set[str]:
    try:
        if active_only:
            return set(
                symbol
                for symbol in client.symbols
                if client.markets.get(symbol, {}).get(
                    enums.ExchangeConstantsMarketStatusColumns.ACTIVE.value, True
                ) in (True, None)
            )
        return set(client.symbols)
    except (AttributeError, TypeError):
        # ccxt exchange load_markets failed
        return set()


def get_time_frames(client):
    try:
        if isinstance(client, ccxt_pro.Exchange):
            # ccxt pro exchanges might have different timeframes in options
            options_time_frames = client.safe_value(client.options, 'timeframes')
            if options_time_frames:
                values = set([
                    time_frame
                    for time_frame in options_time_frames
                    if time_frame_manager.is_time_frame(time_frame)
                ])
                if values:
                    return values
        # use normal client timeframes (values of rest exchange)
        return set(client.timeframes)
    except (AttributeError, TypeError):
        # ccxt exchange describe() is invalid
        return set()


def get_exchange_pair(client, pair) -> str:
    if pair in client.symbols:
        try:
            return client.market(pair)["id"]
        except KeyError:
            pass
    raise ValueError(f'{pair} is not supported')


def get_pair_cryptocurrency(client, pair) -> str:
    if pair in client.symbols:
        try:
            return client.market(pair)["base"]
        except KeyError:
            pass
    raise ValueError(f'{pair} is not supported')


def get_contract_size(client, pair) -> float:
    return get_market_status_contract_size(client.markets[pair])


def get_market_status_contract_size(market_status: dict) :
    return market_status[ccxt_enums.ExchangeConstantsMarketStatusCCXTColumns.CONTRACT_SIZE.value]


def get_fees(market_status) -> dict:
    return {
        enums.ExchangeConstantsMarketPropertyColumns.TAKER.value:
            market_status.get(enums.ExchangeConstantsMarketPropertyColumns.TAKER.value,
                              constants.CONFIG_DEFAULT_FEES),
        enums.ExchangeConstantsMarketPropertyColumns.MAKER.value:
            market_status.get(enums.ExchangeConstantsMarketPropertyColumns.MAKER.value,
                              constants.CONFIG_DEFAULT_FEES),
        enums.ExchangeConstantsMarketPropertyColumns.FEE.value:
            market_status.get(enums.ExchangeConstantsMarketPropertyColumns.FEE.value,
                              constants.CONFIG_DEFAULT_FEES)
    }


def add_headers(client, headers_dict):
    """
    Add new headers to ccxt client
    :param headers_dict: the additional header keys and values as dict
    """
    for header_key, header_value in headers_dict.items():
        client.headers[header_key] = header_value


def add_options(client, options_dict):
    """
    Add new options to ccxt client
    :param options_dict: the additional option keys and values as dict
    """
    for option_key, option_value in options_dict.items():
        client.options[option_key] = option_value


def converted_ccxt_common_errors(f):
    async def converted_ccxt_common_errors_wrapper(*args, **kwargs):
        try:
            return await f(*args, **kwargs)
        except ccxt.RateLimitExceeded as err:
            raise errors.RateLimitExceeded(err) from err
        except ccxt.NotSupported as err:
            raise errors.NotSupported(err) from err
    return converted_ccxt_common_errors_wrapper


def _use_proxy_if_necessary(client, proxy_config: proxy_config_import.ProxyConfig):
    client.aiohttp_trust_env = proxy_config.aiohttp_trust_env
    if proxy_config.http_proxy:
        client.http_proxy = proxy_config.http_proxy
    if proxy_config.http_proxy_callback:
        client.http_proxy_callback = proxy_config.http_proxy_callback
    if proxy_config.https_proxy:
        client.https_proxy = proxy_config.https_proxy
    if proxy_config.https_proxy_callback:
        client.https_proxy_callback = proxy_config.https_proxy_callback
    if proxy_config.socks_proxy:
        client.socks_proxy = proxy_config.socks_proxy
    if proxy_config.socks_proxy_callback:
        client.socks_proxy_callback = proxy_config.socks_proxy_callback


def _get_client_config(
    exchange_class, options, headers, additional_config,
    api_key=None, secret=None, password=None, uid=None,
    auth_token=None, auth_token_header_prefix=None
):
    if auth_token:
        headers["Authorization"] = f"{auth_token_header_prefix or ''}{auth_token}"
    config = {
        'verbose': constants.ENABLE_CCXT_VERBOSE,
        'enableRateLimit': constants.ENABLE_CCXT_RATE_LIMIT,
        'timeout': constants.DEFAULT_REQUEST_TIMEOUT,
        'options': options,
        'headers': headers
    }
    if api_key is not None:
        config['apiKey'] = api_key
    if secret is not None:
        config['secret'] = secret
    if password is not None:
        config['password'] = password
    if uid is not None:
        config['uid'] = uid
    config.update({**_get_custom_domain_config(exchange_class), **(additional_config or {})})
    return config


def _get_custom_domain_config(exchange_class):
    old, new = _get_replaced_custom_domains(exchange_class)
    if not (old and new):
        return {}
    if url_config := exchange_class().describe()[ccxt_enums.ExchangeColumns.URLS.value]:
        commons_logging.get_logger(__name__).info(
            f"Using custom domain for {exchange_class.__name__}: {old} is replaced by {new}"
        )
        return {
            ccxt_enums.ExchangeColumns.URLS.value: _get_patched_url_config(url_config, old, new)
        }
    return {}


def _get_replaced_custom_domains(exchange_class):
    identifier = exchange_class.__name__.upper()
    if custom_domain := os.getenv(f"{identifier}_CUSTOM_DOMAIN"):
        split = custom_domain.split(":")
        if len(split) == 2:
            return split[0], split[1]
        else:
            commons_logging.get_logger(__name__).error(
                f"Invalid {identifier} custom domain config. Expected syntax is to_replace_domain:updated_domain "
                f"Example: MEXC_CUSTOM_DOMAIN=mexc.com:mexc.co"
            )
    return None, None


def _get_patched_url_config(url_config: dict, old: str, new: str):
    updated_config = {}
    for key, val in url_config.items():
        if isinstance(val, dict):
            updated_config[key] = _get_patched_url_config(val, old, new)
        elif isinstance(val, str):
            updated_config[key] = val.replace(old, new)
        else:
            updated_config[key] = val
    return updated_config


def _use_request_counter(identifier: str, ccxt_client: async_ccxt.Exchange):
    """
    Replaces the given exchange async session by an aiohttp_util.CounterClientSession
    WARNING: should only be called right after creating the exchange and on the same async loop as
    the one the exchange will be using (to avoid interrupting open requests from current session and loop errors)
    """
    # use session with request counter
    try:
        # 1. create ssl context and other required elements if necessary
        ccxt_client.open()
        previous_session = ccxt_client.session
        # 2. create patched session using the same params as a normal one
        # same as in ccxt.async_support.exchange.py#open()
        # connector = aiohttp.TCPConnector(ssl=self.ssl_context, loop=self.asyncio_loop, enable_cleanup_closed=True)
        new_connector = aiohttp.TCPConnector(
            ssl=ccxt_client.ssl_context, loop=ccxt_client.asyncio_loop, enable_cleanup_closed=True
        )
        counter_session = aiohttp_util.CounterClientSession(
            identifier,
            loop=ccxt_client.asyncio_loop,
            connector=new_connector,
            trust_env=previous_session.trust_env,
        )
        # 3. replace session
        ccxt_client.session = counter_session
        # 4. close replaced session in task to avoid making this function a coroutine
        asyncio.create_task(previous_session.close())
        commons_logging.get_logger(__name__).info(f"Request counter enabled for {identifier}")
    except Exception as err:
        commons_logging.get_logger(__name__).exception(
            err, True, f"Error when initializing {identifier} request counter: {err}"
        )


def ccxt_exchange_class_factory(exchange_name):
    return getattr(async_ccxt, exchange_name)

def reraise_with_proxy_prefix_if_relevant(
    ccxt_connector, cause_error: Exception, raised_error: typing.Optional[Exception]
):
    was_proxied, last_proxied_request_url = was_latest_request_proxied(ccxt_connector)
    if was_proxied:
        raised = raised_error or cause_error
        raise raised.__class__(f"[Proxied] {raised} [URL: {last_proxied_request_url}]") from cause_error


def was_latest_request_proxied(ccxt_connector) -> (bool, str):
    if not (
        ccxt_connector.exchange_manager.proxy_config
        and ccxt_connector.exchange_manager.proxy_config.get_last_proxied_request_url
    ):
        return False, ""
    last_proxied_request_url = ccxt_connector.exchange_manager.proxy_config.get_last_proxied_request_url()
    last_client_request_url = ccxt_connector.client.last_request_url
    # if last requests are matching: it was proxied
    url_without_param = last_proxied_request_url.split("?")[0]
    return last_proxied_request_url == last_client_request_url, url_without_param


def get_proxy_error_if_any(ccxt_connector, error: Exception) -> typing.Optional[Exception]:
    if not ccxt_connector.exchange_manager.proxy_config:
        return None
    max_depth = 10
    depth = 1
    cause_error = error
    while cause_error and depth < max_depth:
        if isinstance(cause_error, (aiohttp.ClientProxyConnectionError, aiohttp.ClientHttpProxyError)) or (
            isinstance(cause_error, aiohttp.ClientConnectorError)
            and ccxt_connector.exchange_manager.proxy_config.proxy_host in str(cause_error)
        ):
            return cause_error
        depth += 1
        cause_error = getattr(cause_error, "__cause__", None)
    return None



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import mock

import octobot_trading.exchanges as exchanges
import octobot_trading.enums as enums
import pytest

from tests.exchanges import liquid_exchange_manager as liquid_exchange_manager_fixture, DEFAULT_EXCHANGE_NAME
from tests import skipped_on_github_CI

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class FeedCallback:
    def __init__(self, connector, feed):
        self.connector = connector
        self.feed = feed

    def __call__(self, *args, **kwargs):
        if not self.connector.called_feed_event[self.feed].is_set():
            self.connector.called_feed_event[self.feed].set()
            # debug print, start pytest with -s arg to see it in real time
            print(f"Set {self.feed} event from {kwargs}")


class MockedCCXTWebsocketConnector(exchanges.CCXTWebsocketConnector):
    IGNORED_FEED_PAIRS = {
        # Do not ignore Trades feed as it is required for the test
        # When candles are available : use min timeframe kline to push ticker
        enums.WebsocketFeeds.TICKER: [enums.WebsocketFeeds.KLINE]
    }
    EXCHANGE_FEEDS = {
        enums.WebsocketFeeds.CANDLE: True,
        enums.WebsocketFeeds.TICKER: True,
        enums.WebsocketFeeds.TRADES: True,
        enums.WebsocketFeeds.L1_BOOK: True,
    }

    def __init__(self, config, exchange_manager, adapter_class=None, additional_config=None, websocket_name=None):
        super().__init__(config, exchange_manager, adapter_class=adapter_class, additional_config=additional_config,
                         websocket_name=websocket_name)
        self.called_feed_event = {
            feed: asyncio.Event()
            for feed in self.EXCHANGE_FEEDS
        }
        self.callback_mocks = {
            feed: mock.Mock(side_effect=FeedCallback(self, feed))
            for feed in enums.WebsocketFeeds
        }

    async def await_each_feed_call(self, timeout):
        await asyncio.gather(
            *(
                asyncio.wait_for(event.wait(), timeout)
                for event in self.called_feed_event.values()
            )
        )

    async def ticker(self, ticker: dict, symbol=None, **kwargs):
        self.callback_mocks[enums.WebsocketFeeds.TICKER](ticker, symbol=symbol, **kwargs)

    async def recent_trades(self, trades: list, symbol=None, **kwargs):
        self.callback_mocks[enums.WebsocketFeeds.TRADES](trades, symbol=symbol, **kwargs)

    async def book(self, order_book: dict, symbol=None, **kwargs):
        self.callback_mocks[enums.WebsocketFeeds.L1_BOOK](order_book, symbol=symbol, **kwargs)

    async def candle(self, candles: list, symbol=None, timeframe=None, **kwargs):
        self.callback_mocks[enums.WebsocketFeeds.CANDLE](candles, symbol=symbol, timeframe=timeframe, **kwargs)

    async def funding(self, funding: dict, symbol=None, **kwargs):
        self.callback_mocks[enums.WebsocketFeeds.FUNDING](funding, symbol=symbol, **kwargs)

    async def open_interest(self, open_interest: dict, symbol=None, **kwargs):
        self.callback_mocks[enums.WebsocketFeeds.OPEN_INTEREST](open_interest, symbol=symbol, **kwargs)

    async def index(self, index: dict, symbol=None, **kwargs):
        self.callback_mocks[enums.WebsocketFeeds.FUTURES_INDEX](index, symbol=symbol, **kwargs)

    async def orders(self, orders: list, **kwargs):
        self.callback_mocks[enums.WebsocketFeeds.ORDERS](orders, **kwargs)

    async def trades(self, trades: list, **kwargs):
        self.callback_mocks[enums.WebsocketFeeds.TRADE](trades, **kwargs)

    async def balance(self, balance: dict, **kwargs):
        self.callback_mocks[enums.WebsocketFeeds.PORTFOLIO](balance, **kwargs)

    async def transaction(self, transaction: dict, **kwargs):
        self.callback_mocks[enums.WebsocketFeeds.TRANSACTIONS](transaction, **kwargs)

    @classmethod
    def get_name(cls):
        return DEFAULT_EXCHANGE_NAME


class MockedWebSocketExchange(exchanges.DefaultWebSocketExchange):
    DEFAULT_CONNECTOR_CLASS = MockedCCXTWebsocketConnector

    @classmethod
    def get_exchange_connector_class(cls, exchange_manager):
        return cls.DEFAULT_CONNECTOR_CLASS


@pytest.fixture
def default_websocket_exchange(liquid_exchange_manager_fixture):
    yield MockedWebSocketExchange(liquid_exchange_manager_fixture.config, liquid_exchange_manager_fixture)


async def test_start_receive_feeds_and_stop(default_websocket_exchange, skipped_on_github_CI):
    await default_websocket_exchange.init_websocket(
        default_websocket_exchange.exchange_manager.exchange_config.traded_time_frames,
        default_websocket_exchange.exchange_manager.exchange_config.traded_symbol_pairs,
        default_websocket_exchange.exchange_manager.tentacles_setup_config
    )

    # usually last about 5s
    data_reception_timeout = 90
    try:
        await default_websocket_exchange.start_sockets()
        assert len(default_websocket_exchange.websocket_connectors[0].channels) == 4
        await default_websocket_exchange.websocket_connectors[0].await_each_feed_call(data_reception_timeout)
    finally:
        await default_websocket_exchange.stop_sockets()
        await default_websocket_exchange.close_sockets()
        default_websocket_exchange.clear()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums


def parse_position_status(raw_position):
    try:
        return enums.PositionStatus(raw_position[enums.ExchangeConstantsPositionColumns.STATUS.value])
    except KeyError:
        return None


def parse_position_side(raw_position):
    try:
        return enums.PositionSide(raw_position[enums.ExchangeConstantsPositionColumns.SIDE.value])
    except KeyError:
        return None


def parse_position_margin_type(raw_position):
    try:
        return enums.MarginType(raw_position[enums.ExchangeConstantsPositionColumns.MARGIN_TYPE.value])
    except KeyError:
        return None


def parse_position_mode(raw_position):
    try:
        return enums.PositionMode(raw_position[enums.ExchangeConstantsPositionColumns.POSITION_MODE.value])
    except KeyError:
        return None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.orders.types.limit import limit_order
from octobot_trading.personal_data.orders.types.limit.limit_order import (
    LimitOrder,
)

from octobot_trading.personal_data.orders.types.limit import stop_loss_order
from octobot_trading.personal_data.orders.types.limit.stop_loss_order import (
    StopLossOrder,
)

from octobot_trading.personal_data.orders.types.limit import take_profit_order
from octobot_trading.personal_data.orders.types.limit.take_profit_order import (
    TakeProfitOrder,
)

from octobot_trading.personal_data.orders.types.limit import buy_limit_order
from octobot_trading.personal_data.orders.types.limit import sell_limit_order
from octobot_trading.personal_data.orders.types.limit import stop_loss_limit_order
from octobot_trading.personal_data.orders.types.limit import take_profit_limit_order

from octobot_trading.personal_data.orders.types.limit.buy_limit_order import (
    BuyLimitOrder,
)
from octobot_trading.personal_data.orders.types.limit.sell_limit_order import (
    SellLimitOrder,
)
from octobot_trading.personal_data.orders.types.limit.stop_loss_limit_order import (
    StopLossLimitOrder,
)
from octobot_trading.personal_data.orders.types.limit.take_profit_limit_order import (
    TakeProfitLimitOrder,
)

__all__ = [
    "BuyLimitOrder",
    "SellLimitOrder",
    "LimitOrder",
    "TakeProfitOrder",
    "StopLossOrder",
    "StopLossLimitOrder",
    "TakeProfitLimitOrder",
]



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import contextlib

import octobot_commons.html_util as html_util

import octobot_trading.errors as errors
import octobot_trading.constants as constants
import octobot_trading.exchange_data.ticker.channel.ticker as ticker_channel
import octobot_trading.enums as enums
import octobot_trading.exchanges.exchange_websocket_factory as exchange_websocket_factory


class TickerUpdater(ticker_channel.TickerProducer):
    CHANNEL_NAME = constants.TICKER_CHANNEL
    TICKER_REFRESH_TIME = 64
    TICKER_FUTURE_REFRESH_TIME = 14
    TICKER_REFRESH_DELAY_THRESHOLD = 10

    def __init__(self, channel):
        super().__init__(channel)
        self._added_pairs = []
        self.is_fetching_future_data = False
        self.refresh_time = self.TICKER_REFRESH_TIME
        self.updating_pairs = set()

    async def start(self):
        use_futures = self._should_use_future()
        if use_futures:
            self.is_fetching_future_data = True
            self.refresh_time = self.TICKER_FUTURE_REFRESH_TIME
        if self.channel.is_paused:
            await self.pause()
        else:
            if use_futures or self._should_loop():
                # initialize ticker
                await asyncio.gather(*[self._fetch_ticker(pair)
                                       for pair in self._get_pairs_to_update()])
                await asyncio.sleep(self.refresh_time)
                await self.start_update_loop()
            else:
                self.logger.debug(
                    f"Ticker update loop disabled: update is managed by websocket. "
                    f"Updater remains available for forced updates."
                )

    async def start_update_loop(self):
        while not self.should_stop and not self.channel.is_paused:
            try:
                for pair in self._get_pairs_to_update():
                    await self._fetch_ticker(pair)

                await asyncio.sleep(self.refresh_time)
            except errors.NotSupported:
                self.logger.warning(f"{self.channel.exchange_manager.exchange_name} is not supporting updates")
                await self.pause()
            except Exception as e:
                self.logger.exception(
                    e,
                    True,
                    f"Fail to update ticker : {html_util.get_html_summary_if_relevant(e)}"
                )

    async def _fetch_ticker(self, pair):
        try:
            await self.fetch_and_push_pair(pair)
        except errors.FailedRequest as e:
            self.logger.warning(html_util.get_html_summary_if_relevant(e))
            # avoid spamming on disconnected situation
            await asyncio.sleep(constants.FAILED_NETWORK_REQUEST_RETRY_ATTEMPTS)

    async def fetch_and_push_pair(self, pair: str):
        with self._single_pair_update(pair) as can_update:
            if can_update:
                ticker: dict = await self.channel.exchange_manager.exchange.get_price_ticker(pair)
                if self._is_valid(ticker):
                    await self.push(pair, ticker)
                else:
                    self.logger.debug(f"Ignored incomplete ticker: {ticker}")
            else:
                self.logger.debug(f"Skipping {pair} ticker update request: an update is already processing")

    async def trigger_ticker_update(self, symbol: str):
        self.logger.debug(f"Triggered ticker update for {symbol}")
        await self.fetch_and_push_pair(symbol)

    @contextlib.contextmanager
    def _single_pair_update(self, pair: str):
        can_update = False
        try:
            can_update = pair not in self.updating_pairs
            if can_update:
                self.updating_pairs.add(pair)
            yield can_update
        finally:
            if can_update:
                self.updating_pairs.remove(pair)

    @staticmethod
    def _is_valid(ticker):
        try:
            # at least require close, volume and timestamp
            return (
                ticker and all(
                    ticker[field] is not None
                    for field in (
                        enums.ExchangeConstantsTickersColumns.CLOSE.value,
                        enums.ExchangeConstantsTickersColumns.BASE_VOLUME.value,
                        enums.ExchangeConstantsTickersColumns.TIMESTAMP.value
                    )
                )
            )

        except KeyError:
            return False

    def _get_pairs_to_update(self):
        return self.channel.exchange_manager.exchange_config.traded_symbol_pairs + self._added_pairs


    def _should_loop(self):
        """
        Loop when websocket ticker channel not available
        """
        return not exchange_websocket_factory.is_channel_managed_by_websocket(
            self.channel.exchange_manager, constants.TICKER_CHANNEL
        )

    """
    Future data management
    """

    def _should_use_future(self):
        return (
            self.channel.exchange_manager.is_future and (
                self.channel.exchange_manager.exchange.FUNDING_IN_TICKER
                or self.channel.exchange_manager.exchange.MARK_PRICE_IN_TICKER
            )
        )

    async def modify(self, added_pairs=None, removed_pairs=None):
        if added_pairs:
            to_add_pairs = [pair
                            for pair in added_pairs
                            if pair not in self._get_pairs_to_update()]
            if to_add_pairs:
                self._added_pairs += to_add_pairs
                self.logger.info(f"Added pairs : {to_add_pairs}")
                self._update_refresh_time()

        if removed_pairs:
            self._added_pairs -= removed_pairs
            self.logger.info(f"Removed pairs : {removed_pairs}")
            self._update_refresh_time()

    def _update_refresh_time(self):
        if self.is_fetching_future_data:
            # do not change ticker update rate on futures
            return
        pairs_to_update_count = len(self._get_pairs_to_update())
        delay_multiplier = pairs_to_update_count // self.TICKER_REFRESH_DELAY_THRESHOLD + 1
        # there can be many ticker requests when a large number of currency is in a
        # portfolio, in this case, limit those requests
        self.refresh_time = self.TICKER_REFRESH_TIME * delay_multiplier

    async def resume(self) -> None:
        await super().resume()
        if not self.is_running:
            await self.run()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


from octobot_trading.signals import trading_signal_bundle_builder
from octobot_trading.signals.trading_signal_bundle_builder import (
    TradingSignalBundleBuilder,
)
from octobot_trading.signals import util
from octobot_trading.signals.util import (
    create_order_signal_content,
)
from octobot_commons.signals.signal_publisher import (
    SignalPublisher,
)
from octobot_trading.signals import channel
from octobot_trading.signals.channel import (
    RemoteTradingSignalsChannel,
    RemoteTradingSignalChannelProducer,
    RemoteTradingSignalChannelConsumer,
    create_remote_trading_signal_channel_if_missing,
    RemoteTradingSignalProducer,
)
from octobot_trading.signals.signal_creation import (
    remote_signal_publisher,
    should_emit_trading_signal,
    create_order,
    cancel_order,
    edit_order,
    set_leverage,
)


__all__ = [
    "TradingSignalBundleBuilder",
    "create_order_signal_content",
    "SignalPublisher",
    "RemoteTradingSignalsChannel",
    "RemoteTradingSignalChannelProducer",
    "RemoteTradingSignalChannelConsumer",
    "create_remote_trading_signal_channel_if_missing",
    "RemoteTradingSignalProducer",
    "remote_signal_publisher",
    "should_emit_trading_signal",
    "create_order",
    "cancel_order",
    "edit_order",
    "set_leverage",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import enum

import async_channel.channels as channel_instances
import octobot_commons.channels_name as channels_name
import octobot_commons.logging as logging
import octobot_commons.enums as enums
import octobot_commons.constants as commons_constants
import octobot_commons.errors as commons_errors

import octobot_trading.errors as errors
import octobot_trading.exchanges as exchanges
import octobot_trading.modes as modes
import octobot_trading.util as util

OCTOBOT_CHANNEL_TRADING_CONSUMER_LOGGER_TAG = "OctoBotChannelTradingConsumer"


class OctoBotChannelTradingActions(enum.Enum):
    """
    OctoBot Channel consumer supported actions
    """

    EXCHANGE = "exchange"


class OctoBotChannelTradingDataKeys(enum.Enum):
    """
    OctoBot Channel consumer supported data keys
    """

    EXCHANGE_NAME = "exchange_name"
    EXCHANGE_CONFIG = "exchange_config"
    EXCHANGE_ID = "exchange_id"
    BACKTESTING = "backtesting"
    MATRIX_ID = "matrix_id"
    TENTACLES_SETUP_CONFIG = "tentacles_setup_config"


async def octobot_channel_callback(bot_id, subject, action, data) -> None:
    """
    OctoBot channel consumer callback
    :param bot_id: the callback bot id
    :param subject: the callback subject
    :param action: the callback action
    :param data: the callback data
    """
    if subject == enums.OctoBotChannelSubjects.CREATION.value:
        await _handle_creation(bot_id, action, data)


async def _handle_creation(bot_id, action, data):
    if action == OctoBotChannelTradingActions.EXCHANGE.value:
        exchange_name = data.get(OctoBotChannelTradingDataKeys.EXCHANGE_NAME.value, None)
        try:
            config = data[OctoBotChannelTradingDataKeys.EXCHANGE_CONFIG.value]
            exchange_builder = exchanges.create_exchange_builder_instance(config, exchange_name) \
                .has_matrix(data[OctoBotChannelTradingDataKeys.MATRIX_ID.value]) \
                .use_tentacles_setup_config(data[OctoBotChannelTradingDataKeys.TENTACLES_SETUP_CONFIG.value]) \
                .set_bot_id(bot_id)
            try:
                modes.get_activated_trading_mode(data[OctoBotChannelTradingDataKeys.TENTACLES_SETUP_CONFIG.value])
            except commons_errors.ConfigTradingError:
                logging.get_logger(OCTOBOT_CHANNEL_TRADING_CONSUMER_LOGGER_TAG).error(
                    f"No configured trading mode. in order to trade, selected a profile with a trading mode."
                )
                # do not raise on missing trading mode
                exchange_builder.disable_trading_mode()
            _set_exchange_type_details(exchange_builder, config, data[OctoBotChannelTradingDataKeys.BACKTESTING.value])
            await exchange_builder.build()
            await channel_instances.get_chan_at_id(
                channels_name.OctoBotChannelsName.OCTOBOT_CHANNEL.value, bot_id
            ).get_internal_producer().send(
                bot_id=bot_id,
                subject=enums.OctoBotChannelSubjects.NOTIFICATION.value,
                action=action,
                data={OctoBotChannelTradingDataKeys.EXCHANGE_ID.value: exchange_builder.exchange_manager.id}
            )
        except errors.TradingModeIncompatibility as e:
            logging.get_logger(OCTOBOT_CHANNEL_TRADING_CONSUMER_LOGGER_TAG).error(
                f"Error when initializing trading mode, {exchange_name} "
                f"exchange connection is closed to increase performances: {e}")
        except errors.UnreachableExchange as e:
            logging.get_logger(OCTOBOT_CHANNEL_TRADING_CONSUMER_LOGGER_TAG).exception(
                e,
                True,
                f"Error when connecting to {exchange_name} exchange, please check your internet connection ({e})."
            )
        except errors.NotSupported as e:
            logging.get_logger(OCTOBOT_CHANNEL_TRADING_CONSUMER_LOGGER_TAG).exception(e, True, str(e))
        except Exception as e:
            logging.get_logger(OCTOBOT_CHANNEL_TRADING_CONSUMER_LOGGER_TAG).exception(
                e,
                True,
                f"Error when creating a new {exchange_name} exchange connexion: {e.__class__.__name__} {e}"
            )


def _set_exchange_type_details(exchange_builder, config, backtesting):
    # real, simulator, backtesting
    if util.is_trader_enabled(config):
        exchange_builder.is_real()
    elif util.is_trader_simulator_enabled(config):
        exchange_builder.is_simulated()
    if backtesting is not None:
        exchange_builder.is_simulated()
        exchange_builder.is_rest_only()
        exchange_builder.is_backtesting(backtesting)
    # use exchange sandbox
    exchange_builder.is_sandboxed(
        config[commons_constants.CONFIG_EXCHANGES].get(exchange_builder.exchange_name, {}).get(
            commons_constants.CONFIG_EXCHANGE_SANDBOXED, False)
    )
    # exchange trading type
    config_exchange_type = config[commons_constants.CONFIG_EXCHANGES].get(exchange_builder.exchange_name, {}).get(
        commons_constants.CONFIG_EXCHANGE_TYPE, exchanges.get_default_exchange_type(exchange_builder.exchange_name))
    exchange_builder.is_using_exchange_type(config_exchange_type)

    # rest, web socket
    if config[commons_constants.CONFIG_EXCHANGES].get(exchange_builder.exchange_name, {}).get(
            commons_constants.CONFIG_EXCHANGE_REST_ONLY, False):
        exchange_builder.is_rest_only()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import os
import asyncio

import mock
import pytest
from mock import patch, Mock, AsyncMock
import octobot_commons.constants as commons_constants
import octobot_trading.personal_data as personal_data
import octobot_trading.constants as constants
from octobot_trading.personal_data import SellLimitOrder
from octobot_trading.personal_data.orders import BuyMarketOrder, BuyLimitOrder

from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests import event_loop

# All test coroutines will be treated as marked.
from tests.personal_data import DEFAULT_MARKET_QUANTITY
from tests.test_utils.random_numbers import decimal_random_price, decimal_random_quantity

pytestmark = pytest.mark.asyncio


async def test_handle_balance_update(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    if os.getenv('CYTHON_IGNORE'):
        return

    with patch.object(portfolio_manager.portfolio, 'update_portfolio_from_balance',
                      new=Mock()) as update_portfolio_from_balance_mock:
        update_portfolio_from_balance_mock.assert_not_called()

        portfolio_manager.handle_balance_update(None)
        update_portfolio_from_balance_mock.assert_not_called()

        trader.is_enabled = False
        portfolio_manager.handle_balance_update({})
        update_portfolio_from_balance_mock.assert_not_called()

        trader.is_enabled = True
        portfolio_manager.handle_balance_update({})
        update_portfolio_from_balance_mock.assert_called_once()


async def test_handle_balance_update_from_order(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    trader.simulate = False
    order = BuyMarketOrder(trader)

    with patch.object(portfolio_manager, '_refresh_real_trader_portfolio_until_order_update_applied',
                      new=AsyncMock()) as _refresh_real_trader_portfolio_until_order_update_applied_mock, \
        patch.object(portfolio_manager, '_refresh_simulated_trader_portfolio_from_order',
                     new=Mock()) as _refresh_simulated_trader_portfolio_from_order_mock:
        _refresh_real_trader_portfolio_until_order_update_applied_mock.assert_not_called()
        await portfolio_manager.handle_balance_update_from_order(order, True, False)
        _refresh_real_trader_portfolio_until_order_update_applied_mock.assert_called_once_with(order, False)
        _refresh_simulated_trader_portfolio_from_order_mock.assert_not_called()
        _refresh_real_trader_portfolio_until_order_update_applied_mock.reset_mock()
        await portfolio_manager.handle_balance_update_from_order(order, True, True)
        _refresh_real_trader_portfolio_until_order_update_applied_mock.assert_called_once_with(order, True)
        _refresh_simulated_trader_portfolio_from_order_mock.assert_not_called()
        _refresh_real_trader_portfolio_until_order_update_applied_mock.reset_mock()
        with portfolio_manager.disabled_portfolio_update_from_order():
            await portfolio_manager.handle_balance_update_from_order(order, False, False)
            _refresh_real_trader_portfolio_until_order_update_applied_mock.assert_not_called()
            _refresh_simulated_trader_portfolio_from_order_mock.assert_not_called()
            await portfolio_manager.handle_balance_update_from_order(order, True, False)
            _refresh_real_trader_portfolio_until_order_update_applied_mock.assert_not_called()
            _refresh_simulated_trader_portfolio_from_order_mock.assert_not_called()
        await portfolio_manager.handle_balance_update_from_order(order, False, False)
        _refresh_real_trader_portfolio_until_order_update_applied_mock.assert_not_called()
        _refresh_simulated_trader_portfolio_from_order_mock.assert_called_once()

    trader.simulate = True
    with patch.object(portfolio_manager, '_refresh_simulated_trader_portfolio_from_order',
                      new=Mock()) as _refresh_simulated_trader_portfolio_from_order_mock:
        _refresh_simulated_trader_portfolio_from_order_mock.assert_not_called()
        await portfolio_manager.handle_balance_update_from_order(order, True, False)
        _refresh_simulated_trader_portfolio_from_order_mock.assert_called_once()
        _refresh_simulated_trader_portfolio_from_order_mock.reset_mock()
        with portfolio_manager.disabled_portfolio_update_from_order():
            await portfolio_manager.handle_balance_update_from_order(order, True, False)
            _refresh_simulated_trader_portfolio_from_order_mock.assert_not_called()
            await portfolio_manager.handle_balance_update_from_order(order, False, False)
            _refresh_simulated_trader_portfolio_from_order_mock.assert_not_called()
        # ensure no side effect with require_exchange_update param
        await portfolio_manager.handle_balance_update_from_order(order, False, False)
        _refresh_simulated_trader_portfolio_from_order_mock.assert_called_once()

    trader.is_enabled = False
    trader.simulate = False
    assert not await portfolio_manager.handle_balance_update_from_order(order, True, False)
    assert not await portfolio_manager.handle_balance_update_from_order(order, False, False)


async def test_refresh_real_trader_portfolio_until_order_update_applied(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    trader.simulate = False
    order = BuyMarketOrder(trader)
    with patch.object(portfolio_manager, '_refresh_real_trader_portfolio',
                      new=AsyncMock(return_value=True)) as _refresh_real_trader_portfolio_mock, \
        patch.object(asyncio, 'sleep', AsyncMock()) as sleep_mock:
        with patch.object(portfolio_manager, '_has_filled_order_been_applied',
                          new=Mock(return_value=True)) as _has_filled_order_been_applied_mock:
            assert await portfolio_manager._refresh_real_trader_portfolio_until_order_update_applied(order, False) is True
            _has_filled_order_been_applied_mock.assert_not_called()
            _refresh_real_trader_portfolio_mock.assert_called_once()
            sleep_mock.assert_not_called()
            _refresh_real_trader_portfolio_mock.reset_mock()
            assert await portfolio_manager._refresh_real_trader_portfolio_until_order_update_applied(order, True) is True
            _has_filled_order_been_applied_mock.assert_called_once_with(order)
            sleep_mock.assert_not_called()
            _refresh_real_trader_portfolio_mock.reset_mock()
        with patch.object(portfolio_manager, '_has_filled_order_been_applied',
                          new=Mock(return_value=False)) as _has_filled_order_been_applied_mock:
            assert await portfolio_manager._refresh_real_trader_portfolio_until_order_update_applied(order, True) is True
            assert _has_filled_order_been_applied_mock.call_count == 1 + constants.MAX_PORTFOLIO_SYNC_ATTEMPTS
            assert _refresh_real_trader_portfolio_mock.call_count == 1 + constants.MAX_PORTFOLIO_SYNC_ATTEMPTS
            assert sleep_mock.call_count == constants.MAX_PORTFOLIO_SYNC_ATTEMPTS
            _has_filled_order_been_applied_mock.reset_mock()
            _refresh_real_trader_portfolio_mock.reset_mock()
            sleep_mock.reset_mock()
        calls = []

        # retry once
        def _has_filled_order_been_applied(*args):
            if calls:
                return True
            calls.append(1)
            return False
        with patch.object(portfolio_manager, '_has_filled_order_been_applied',
                          new=Mock(side_effect=_has_filled_order_been_applied)) as _has_filled_order_been_applied_mock:
            assert await portfolio_manager._refresh_real_trader_portfolio_until_order_update_applied(order, True) is True
            assert _has_filled_order_been_applied_mock.call_count == 2
            assert _refresh_real_trader_portfolio_mock.call_count == 2
            assert sleep_mock.call_count == 1


async def test_has_filled_order_been_applied(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    trader.simulate = False
    order = BuyMarketOrder(trader)
    order.update("BTC/USDT", quantity=decimal.Decimal(10), price=decimal.Decimal(100))

    # 10 BTC in PF
    assert portfolio_manager._has_filled_order_been_applied(order) is True

    order.update("BTC/USDT", quantity=decimal.Decimal(15), price=decimal.Decimal(100))
    assert portfolio_manager._has_filled_order_been_applied(order) is False

    order = SellLimitOrder(trader)
    order.update("BTC/USDT", quantity=decimal.Decimal(10), price=decimal.Decimal(100))
    # 1000 USDT in PF
    assert portfolio_manager._has_filled_order_been_applied(order) is True

    order.update("BTC/USDT", quantity=decimal.Decimal(10), price=decimal.Decimal(200))
    assert portfolio_manager._has_filled_order_been_applied(order) is False

    trader.exchange_manager.is_future = True
    # always True on futures
    assert portfolio_manager._has_filled_order_been_applied(order) is True
    trader.exchange_manager.is_future = False


async def test_refresh_simulated_trader_portfolio_from_order(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    if os.getenv('CYTHON_IGNORE'):
        return
    order = BuyLimitOrder(trader)
    order.symbol = "BTC/USDT"
    await order.initialize()
    with patch.object(portfolio_manager.portfolio, 'update_portfolio_available',
                      new=Mock()) as update_portfolio_available_mock:
        update_portfolio_available_mock.assert_not_called()
        portfolio_manager._refresh_simulated_trader_portfolio_from_order(order)
        update_portfolio_available_mock.assert_called_once()

    price = decimal_random_price()
    order.update(
        price=decimal_random_price(),
        quantity=decimal_random_quantity(max_value=DEFAULT_MARKET_QUANTITY / price),
        symbol="BTC/USDT"
    )
    await order.on_fill(force_fill=True)
    assert order.is_filled()

    with patch.object(portfolio_manager.portfolio, 'update_portfolio_from_filled_order',
                      new=Mock()) as update_portfolio_from_filled_order_mock:
        update_portfolio_from_filled_order_mock.assert_not_called()
        portfolio_manager._refresh_simulated_trader_portfolio_from_order(order)
        update_portfolio_from_filled_order_mock.assert_called_once()


async def test_load_simulated_portfolio_from_history(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    portfolio_manager.historical_portfolio_value_manager = mock.Mock(
        historical_ending_portfolio={
            "BTC": {
                commons_constants.PORTFOLIO_AVAILABLE: 1,
                commons_constants.PORTFOLIO_TOTAL: 10.11,
            },
            "ETH": {
                commons_constants.PORTFOLIO_AVAILABLE: -1,
                commons_constants.PORTFOLIO_TOTAL: 10,
            },
            "USDT": {
                commons_constants.PORTFOLIO_AVAILABLE: 34,
                commons_constants.PORTFOLIO_TOTAL: 34,
            }
        },
        stop=mock.AsyncMock()
    )
    portfolio_manager._load_simulated_portfolio_from_history()
    # ensure only the total value is loaded in simulated portfolio
    assert portfolio_manager.portfolio.portfolio == {
        "BTC": personal_data.SpotAsset("BTC", decimal.Decimal("10.11"), decimal.Decimal("10.11")),
        "ETH": personal_data.SpotAsset("ETH", decimal.Decimal("10"), decimal.Decimal("10")),
        "USDT": personal_data.SpotAsset("USDT", decimal.Decimal("34"), decimal.Decimal("34"))
    }



# pylint: disable=W0706
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import decimal
import typing

import octobot_commons.logging as logging
import octobot_commons.html_util as html_util
import octobot_commons.constants

import octobot_trading.personal_data.orders.order_factory as order_factory
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.personal_data.orders.decimal_order_adapter as decimal_order_adapter
import octobot_trading.personal_data.trades.trade_factory as trade_factory
import octobot_trading.enums as enums
import octobot_trading.constants
import octobot_trading.errors as errors
import octobot_trading.util as util
import octobot_trading.signals as signals


class Trader(util.Initializable):
    NO_HISTORY_MESSAGE = "Starting a fresh new trading session using the current portfolio as a profitability " \
                         "reference."

    def __init__(self, config, exchange_manager):
        super().__init__()
        self.exchange_manager = exchange_manager
        self.config = config

        self.risk = octobot_trading.constants.ZERO
        try:
            self.set_risk(decimal.Decimal(str(self.config[octobot_commons.constants.CONFIG_TRADING]
                                              [octobot_commons.constants.CONFIG_TRADER_RISK])))
        except KeyError:
            self.set_risk(octobot_trading.constants.ZERO)
        self.allow_artificial_orders = self.config.get(octobot_commons.constants.CONFIG_TRADER_ALLOW_ARTIFICIAL_ORDERS,
                                                       True)

        # logging
        self.trader_type_str = octobot_trading.constants.REAL_TRADER_STR
        self.logger = logging.get_logger(f"{self.__class__.__name__}[{self.exchange_manager.exchange_name}]")

        if not hasattr(self, 'simulate'):
            self.simulate = False
        self.is_enabled = self.__class__.enabled(self.config)

    async def initialize_impl(self):
        self.is_enabled = self.is_enabled and self.exchange_manager.is_trading
        if self.is_enabled:
            await self.exchange_manager.register_trader(self)
        self.logger.debug(f"{'Enabled' if self.is_enabled else 'Disabled'} on {self.exchange_manager.exchange_name}")

    def clear(self):
        self.exchange_manager = None

    @classmethod
    def enabled(cls, config):
        return util.is_trader_enabled(config)

    def set_risk(self, risk):
        min_risk = decimal.Decimal(str(octobot_commons.constants.CONFIG_TRADER_RISK_MIN))
        max_risk = decimal.Decimal(str(octobot_commons.constants.CONFIG_TRADER_RISK_MAX))
        if risk < min_risk:
            self.risk = min_risk
        elif risk > max_risk:
            self.risk = max_risk
        else:
            self.risk = risk
        return self.risk

    """
    Orders
    """

    async def create_order(
        self, order, loaded: bool = False, params: dict = None, wait_for_creation=True, raise_all_creation_error=False,
        creation_timeout=octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
    ):
        """
        Create a new order from an OrderFactory created order, update portfolio, registers order in order manager and
        notifies order channel.
        :param order: Order to create
        :param loaded: True if this order is fetched from an exchange only and therefore not created by this OctoBot
        :param params: Additional parameters to give to the order upon creation (used in real trading only)
        :param wait_for_creation: when True, always make sure the order is completely created before returning.
        On exchanges async api, a create request will return before the order is actually live on exchange, in this case
        the associated order state will make sure that the order is creating by polling the order from the exchange.
        :param raise_all_creation_error: when True, will raise each ceation error when possible
        (instead of retuning None)
        :param creation_timeout: time before raising a timeout error when waiting for an order creation
        :return: The crated order instance
        """
        if loaded:
            order.is_from_this_octobot = False
            self.logger.debug(f"Order loaded : {order.to_string()} ")
            # force initialize to always create open state
            await order.initialize()
            return order
        # octobot order
        created_order = order
        try:
            params = params or {}
            self.logger.info(f"Creating order: {created_order}")
            created_order = await self._create_new_order(order, params, wait_for_creation=wait_for_creation,
                                                         creation_timeout=creation_timeout)
            if created_order is None:
                self.logger.warning(f"Order not created on {self.exchange_manager.exchange_name} "
                                    f"(failed attempt to create: {order}). This is likely due to "
                                    f"the order being refused by the exchange.")
        except (
            errors.MissingFunds, errors.AuthenticationError,
            errors.ExchangeCompliancyError, errors.OrderCreationError
        ):
            # forward errors that require actions to fix the situation
            raise
        except Exception as e:
            if raise_all_creation_error:
                raise
            self.logger.exception(e, True, f"Unexpected error when creating order: {e}. Order: {order}")
            return None

        return created_order

    async def create_artificial_order(
        self, order_type, symbol, current_price, quantity, price, reduce_only, close_position,
        emit_trading_signals=False, wait_for_creation=True,
        creation_timeout=octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
    ):
        """
        Creates an OctoBot managed order (managed orders example: stop loss that is not published on the exchange and
        that is maintained internally).
        """
        order = order_factory.create_order_instance(
            trader=self,
            order_type=order_type,
            symbol=symbol,
            current_price=current_price,
            quantity=quantity,
            price=price,
            reduce_only=reduce_only,
            close_position=close_position,
        )
        async with signals.remote_signal_publisher(self.exchange_manager, order.symbol, emit_trading_signals):
            return await signals.create_order(
                self.exchange_manager,
                emit_trading_signals and signals.should_emit_trading_signal(self.exchange_manager),
                order,
                wait_for_creation=wait_for_creation,
                creation_timeout=creation_timeout
            )

    async def edit_order(self, order,
                         edited_quantity: decimal.Decimal = None,
                         edited_price: decimal.Decimal = None,
                         edited_stop_price: decimal.Decimal = None,
                         edited_current_price: decimal.Decimal = None,
                         params: dict = None) -> bool:
        """
        Edits an order, might be a simulated or a real order.
        Fields that can be edited are:
            quantity, price, stop_price and current_price
        Portfolio is updated within this call
        :return: True when an order field got updated
        """
        if not order.can_be_edited():
            raise errors.OrderEditError(f"Order can't be edited, order: {order}")
        changed = False
        previous_exchange_order_id = order.exchange_order_id
        try:
            async with order.lock:
                # now that we got the lock, ensure we can edit the order
                if not self.simulate and not order.is_self_managed() and order.state is not None:
                    # careful here: make sure we are not editing an order on exchange that is being updated
                    # somewhere else
                    async with order.state.refresh_operation():
                        self.logger.info(f"Editing order: {order} ["
                                         f"edited_quantity: {str(edited_quantity)} "
                                         f"edited_price: {str(edited_price)} "
                                         f"edited_stop_price: {str(edited_stop_price)} "
                                         f"edited_current_price: {str(edited_current_price)}"
                                         f"]")
                        order_params = self.exchange_manager.exchange.get_order_additional_params(order)
                        order_params.update(params or {})
                        # fill in every param as some exchange rely on re-creating the order altogether
                        edited_order = await self.exchange_manager.exchange.edit_order(
                            order.exchange_order_id,
                            order.order_type,
                            order.symbol,
                            quantity=order.origin_quantity if edited_quantity is None else edited_quantity,
                            price=order.origin_price if edited_price is None else edited_price,
                            stop_price=edited_stop_price,
                            side=order.side,
                            current_price=edited_current_price,
                            params=order_params
                        )
                        # apply new values from returned order (even order id might have changed)
                        self.logger.debug(f"Successfully edited order on {self.exchange_manager.exchange_name}, "
                                          f"new order values: {edited_order}")
                        changed = order.update_from_raw(edited_order)
                        # update portfolio from exchange
                        await self.exchange_manager.exchange_personal_data.\
                            handle_portfolio_and_position_update_from_order(order)
                else:
                    # consider order as cancelled to release portfolio amounts
                    self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.update_portfolio_available(
                        order, is_new_order=False
                    )
                    changed = order.update(
                        order.symbol,
                        quantity=edited_quantity,
                        price=edited_stop_price if edited_price is None else edited_price,
                        stop_price=edited_stop_price,
                        current_price=edited_current_price,
                    )
                    # consider order as new order to lock portfolio amounts
                    self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.update_portfolio_available(
                        order, is_new_order=True
                    )
                # push order edit into orders channel as edit update
                await self.exchange_manager.exchange_personal_data.handle_order_update_notification(
                    order, enums.OrderUpdateType.EDIT
                )
                self.logger.info(f"Edited order: {order}")
            return changed
        finally:
            if previous_exchange_order_id != order.exchange_order_id:
                # order id changed: update orders_manager to keep consistency
                self.exchange_manager.exchange_personal_data.orders_manager.replace_order(order.order_id, order)

    async def _create_new_order(self, new_order, params: dict,
                                wait_for_creation=True,
                                creation_timeout=octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT) -> object:
        """
        Creates an exchange managed order, it might be a simulated or a real order.
        Portfolio will be updated by the created order state after order will be initialized
        """
        updated_order = new_order
        is_pending_creation = False
        if not self.simulate and not new_order.is_self_managed():
            order_params = self.exchange_manager.exchange.get_order_additional_params(new_order)
            order_params.update(new_order.exchange_creation_params)
            order_params.update(params)
            created_order = await self.exchange_manager.exchange.create_order(
                order_type=new_order.order_type,
                symbol=new_order.symbol,
                quantity=new_order.origin_quantity,
                price=new_order.origin_price,
                stop_price=new_order.origin_stop_price,
                side=new_order.side,
                current_price=new_order.created_last_price,
                reduce_only=new_order.reduce_only,
                params=order_params
            )
            if created_order is None:
                return None
            self.logger.debug(f"Successfully created order on {self.exchange_manager.exchange_name}: {created_order}")

            # get real order from exchange
            updated_order = order_factory.create_order_instance_from_raw(
                self, created_order, force_open_or_pending_creation=True, has_just_been_created=True
            )
            is_pending_creation = updated_order.status == enums.OrderStatus.PENDING_CREATION

            # rebind local elements to new order instance
            if new_order.order_group:
                updated_order.add_to_order_group(new_order.order_group)
            updated_order.order_id = new_order.order_id
            updated_order.tag = new_order.tag
            updated_order.chained_orders = new_order.chained_orders
            for chained_order in new_order.chained_orders:
                chained_order.triggered_by = updated_order
            updated_order.triggered_by = new_order.triggered_by
            updated_order.has_been_bundled = new_order.has_been_bundled
            updated_order.exchange_creation_params = new_order.exchange_creation_params
            updated_order.is_waiting_for_chained_trigger = new_order.is_waiting_for_chained_trigger
            updated_order.associated_entry_ids = new_order.associated_entry_ids
            updated_order.update_with_triggering_order_fees = new_order.update_with_triggering_order_fees

            if is_pending_creation:
                # register order as pending order, it will then be added to live orders in order manager once open
                self.exchange_manager.exchange_personal_data.orders_manager.register_pending_creation_order(
                    updated_order
                )

        await updated_order.initialize()
        if is_pending_creation and wait_for_creation \
                and updated_order.state is not None and updated_order.state.is_pending()\
                and self.exchange_manager.exchange_personal_data.orders_manager.enable_order_auto_synchronization:
            await updated_order.state.wait_for_terminate(creation_timeout)
        return updated_order

    def get_take_profit_order_type(self, base_order, order_type: enums.TraderOrderType) -> enums.TraderOrderType:
        """
        Returns the adapted take profit order enums.TraderOrderType.
        :return: enums.TraderOrderType.TAKE_PROFIT when order can be bundled and considered as a real take profit
        from exchange and the given order_type otherwise
        """
        if not self.simulate and self.exchange_manager.exchange.supports_bundled_order_on_order_creation(
            base_order, enums.TraderOrderType.TAKE_PROFIT
        ):
            # use take profit order type for bundled orders, use default order type otherwise
            return enums.TraderOrderType.TAKE_PROFIT
        return order_type

    async def bundle_chained_order_with_uncreated_order(
        self, order, chained_order, update_with_triggering_order_fees, **kwargs
    ) -> dict:
        """
        Creates and bundles an order as a chained order to the given order.
        When supported and in real trading, return the stop loss parameters to be given when
        pushing the initial order on exchange
        :param order: the order to create a chained order from after fill
        :param chained_order: the chained order to create when the 1st order is filled
        :param update_with_triggering_order_fees: if the chained order quantity should
        be updated with triggering order fees
        :return: parameters with chained order details if supported
        """
        params = {}
        is_bundled = self.exchange_manager.exchange.supports_bundled_order_on_order_creation(
            order, chained_order.order_type
        )
        if is_bundled:
            # warning: doesn't work for multiple stop loss / take profits
            if chained_order.order_type is enums.TraderOrderType.STOP_LOSS:
                params.update(self.exchange_manager.exchange.get_bundled_order_parameters(
                    order,
                    stop_loss_price=chained_order.origin_price
                ))
            elif chained_order.order_type in (enums.TraderOrderType.TAKE_PROFIT,
                                              enums.TraderOrderType.BUY_MARKET, enums.TraderOrderType.SELL_MARKET,
                                              enums.TraderOrderType.BUY_LIMIT, enums.TraderOrderType.SELL_LIMIT):
                params.update(self.exchange_manager.exchange.get_bundled_order_parameters(
                    order,
                    stop_loss_price=None,   # required for cython
                    take_profit_price=chained_order.origin_price
                ))
            if params:
                self.logger.debug(
                    f"Including {chained_order.order_type} chained order into order "
                    f"parameters to handle it directly on exchange."
                )
        await self.chain_order(order, chained_order, update_with_triggering_order_fees, is_bundled, **kwargs)
        return params

    async def chain_order(self, order, chained_order, update_with_triggering_order_fees, is_bundled, **kwargs):
        await chained_order.set_as_chained_order(order, is_bundled, {}, update_with_triggering_order_fees, **kwargs)
        order.add_chained_order(chained_order)
        self.logger.info(f"Added chained order [{chained_order}] to [{order}] order.")

    async def cancel_order(self, order, ignored_order=None,
                           wait_for_cancelling=True,
                           cancelling_timeout=octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT) -> bool:
        """
        Cancels the given order and updates the portfolio, publish in order channel
        if order is from a real exchange.
        :param order: Order to cancel
        :param ignored_order: Order not to cancel if found in groupped orders recursive cancels (ex: avoid cancelling
        a filled order)
        :param wait_for_cancelling: when True, always make sure the order is completely cancelled before returning.
        On exchanges async api, a cancel request will return before the order is actually cancelled, in this case
        the associated order state will make sure that the order is cancelled by polling the order from the exchange.
        :param cancelling_timeout: time before raising a timeout error when waiting for an order cancel
        :return: None
        """
        if order and order.is_open():
            self.logger.info(f"Cancelling order: {order}")
            # always cancel this order first to avoid infinite loop followed by deadlock
            return await self._handle_order_cancellation(order, ignored_order,
                                                         wait_for_cancelling, cancelling_timeout)
        return False

    async def _handle_order_cancellation(
        self, order, ignored_order, wait_for_cancelling: bool, cancelling_timeout: float
    ) -> bool:
        success = True
        if order.is_waiting_for_chained_trigger:
            # order will just never get created
            order.is_waiting_for_chained_trigger = False
            return success
        order_status = None
        # if real order: cancel on exchange
        if not self.simulate and not order.is_self_managed():
            try:
                async with order.lock:
                    try:
                        order_status = await self.exchange_manager.exchange.cancel_order(
                            order.exchange_order_id, order.symbol, order.order_type
                        )
                    except errors.NotSupported:
                        raise
                    except (errors.OrderCancelError, Exception) as inner_err:
                        # retry to cancel order
                        self.logger.info(
                            f"Failed to cancel order ({inner_err} {inner_err.__class__.__name__}), retrying"
                        )
                        order_status = await self.exchange_manager.exchange.cancel_order(
                            order.exchange_order_id, order.symbol, order.order_type
                        )
            except errors.OrderCancelError as err:
                if self.exchange_manager.exchange_personal_data.orders_manager.enable_order_auto_synchronization:
                    if await self._handle_order_cancel_error(order, err, wait_for_cancelling, cancelling_timeout):
                        return True
                else:
                    self.logger.warning(
                        f"Impossible to cancel order ({err} {err.__class__.__name__}). "
                        f"Considering order as cancelled {order}"
                    )
                    order_status = enums.OrderStatus.CANCELED
            except Exception as err:
                self.logger.exception(err, True, f"Failed to cancel order {order}")
                return False
            if order_status is enums.OrderStatus.CANCELED:
                order.status = octobot_trading.enums.OrderStatus.CANCELED
                self.logger.debug(f"Successfully cancelled order {order}")
            elif order_status is enums.OrderStatus.PENDING_CANCEL:
                order.status = octobot_trading.enums.OrderStatus.PENDING_CANCEL
                self.logger.debug(f"Order cancel in progress for {order}")
        else:
            order.status = octobot_trading.enums.OrderStatus.CANCELED

        await order.on_cancel(force_cancel=order.status is octobot_trading.enums.OrderStatus.CANCELED,
                              is_from_exchange_data=False,
                              ignored_order=ignored_order)
        if wait_for_cancelling and order.state is not None and order.state.is_pending():
            await self._wait_for_order_cancel(order, cancelling_timeout)
        return True

    async def _handle_order_cancel_error(self, order, err, wait_for_cancelling, cancelling_timeout):
        """
        Use when an order can't be cancelled: it usually means that the order is not open on exchange anymore.
        Will synch the given order on exchange to figure out the cancel issue if any.
        Returns True when the order cancel ends up successful even though an error initially occurred
        Raises OrderCancelError on unrecoverable cases
        Raises a subclass of UnexpectedExchangeSideOrderStateError when the order is in a unexpected state
        on exchange but is still manageable.
        """
        if order.state is None:
            raise errors.OrderCancelError(
                f"Error when cancelling order. This order state is unset, which makes "
                f"it impossible to handle this the issue. Please report it if you see it. "
                f"Order: {order}"
            ) from err
        # trigger forced refresh to get an update of the order
        if order.state.is_refreshing():
            await order.state.wait_for_next_state(cancelling_timeout)
        else:
            previous_status = order.status
            await order.state.synchronize(force_synchronization=True)
            if previous_status != order.status:
                # status changed: wait for state change
                await order.state.wait_for_next_state(cancelling_timeout)
        if order.is_cancelled():
            self.logger.debug(f"Tried to cancel an already cancelled order. Order: {order}")
            return True
        if order.is_cancelling():
            if wait_for_cancelling:
                await self._wait_for_order_cancel(order, cancelling_timeout)
            return True
        elif order.is_open():
            if isinstance(err, errors.OrderNotFoundOnCancelError):
                raise errors.OrderNotFoundOnCancelError(
                    f"Tried to cancel an order that can't be found, it might be cancelled or filled already "
                    f"({html_util.get_html_summary_if_relevant(err)}). "
                    f"Order: {order}"
                ) from err
            raise errors.OpenOrderError(
                f"Order is open, but can't be cancelled. This is unexpected. Order: {order}"
            ) from err
        elif order.is_filled():
            raise errors.FilledOrderError(f"Order is filled, it can't be cancelled. Order: {order}") from err
        elif order.is_closed():
            raise errors.ClosedOrderError(f"Order is closed, it can't be cancelled. Order: {order}") from err
        else:
            # should not happen
            raise errors.OrderCancelError(
                f"Can't cancel order and unknown post sync order state for order: {order}."
            ) from err

    async def cancel_all_orders(
        self,
        symbol: str,
        allow_single_order_cancel_fallback: bool,
        wait_for_cancelling: bool = True,
        cancelling_timeout: float = octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
    ) -> bool:
        orders_to_cancel = self.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(symbol)
        success = True
        if orders_to_cancel:
            self.logger.info(f"Cancelling all {len(orders_to_cancel)} {symbol} orders")
        else:
            return success
        try:
            cancel_on_exchange = False
            for order in orders_to_cancel:
                await order.lock.acquire()
                if self.simulate or order.is_self_managed():
                    order.status = octobot_trading.enums.OrderStatus.CANCELED
                else:
                    cancel_on_exchange = True
            if cancel_on_exchange:
                try:
                    success = False
                    await self.exchange_manager.exchange.cancel_all_orders(symbol)
                    success = True
                except errors.NotSupported:
                    if allow_single_order_cancel_fallback:
                        self.logger.debug(
                            f"cancel_all_orders is not supported on {self.exchange_manager.exchange_name}. Falling "
                            f"back to one by one cancel"
                        )
                        for order in orders_to_cancel:
                            if not await self.cancel_order(
                                order.symbol,
                                wait_for_cancelling=wait_for_cancelling,
                                cancelling_timeout=cancelling_timeout
                            ):
                                success = False
                    else:
                        # not supported and no fallback allowed: re-raise errors.NotSupported
                        raise
            for order in orders_to_cancel:
                await order.on_cancel(force_cancel=True, is_from_exchange_data=False)
                if wait_for_cancelling and order.state is not None and order.state.is_pending():
                    await self._wait_for_order_cancel(order, cancelling_timeout)
        finally:
            for order in orders_to_cancel:
                order.lock.release()
        self.logger.info(f"Cancelling of all {len(orders_to_cancel)} {symbol} orders complete")
        return success

    async def _wait_for_order_cancel(self, order, cancelling_timeout):
        self.logger.debug(f"Waiting for order cancelling, order: {order}")
        await order.state.wait_for_terminate(cancelling_timeout)
        self.logger.debug(f"Completed order cancelling, order: {order}")

    async def cancel_order_with_id(self, exchange_order_id, emit_trading_signals=False,
                                   wait_for_cancelling=True,
                                   cancelling_timeout=octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT):
        """
        Gets order matching order_id from the OrderManager and calls self.cancel_order() on it
        :param exchange_order_id: Exchange id of the order to cancel
        :param emit_trading_signals: when true, trading signals will be emitted
        :param wait_for_cancelling: when True, always make sure the order is completely cancelled before returning.
        :param cancelling_timeout: time before raising a timeout error when waiting for an order cancel
        :return: True if cancel is successful, False if order is not found or cancellation failed
        """
        try:
            order = self.exchange_manager.exchange_personal_data.orders_manager.get_order(exchange_order_id)
            async with signals.remote_signal_publisher(self.exchange_manager, order.symbol, emit_trading_signals):
                return await signals.cancel_order(
                    self.exchange_manager,
                    emit_trading_signals and signals.should_emit_trading_signal(self.exchange_manager),
                    order,
                    wait_for_cancelling=wait_for_cancelling,
                    cancelling_timeout=cancelling_timeout,
                )
        except KeyError:
            return False

    async def cancel_open_orders(self, symbol, cancel_loaded_orders=True, side=None,
                                 emit_trading_signals=False,
                                 wait_for_cancelling=True,
                                 cancelling_timeout=octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT, 
                                 since: typing.Union[int, float] = octobot_trading.constants.NO_DATA_LIMIT, 
                                 until: typing.Union[int, float] = octobot_trading.constants.NO_DATA_LIMIT
                                 ) -> (bool, list):
        """
        Should be called only if the goal is to cancel all open orders for a given symbol
        :param symbol: The symbol to cancel all orders on
        :param cancel_loaded_orders: When True, also cancels loaded orders (order that are not from this bot instance)
        :param side: When set, only cancels orders from this side
        :param emit_trading_signals: when true, trading signals will be emitted
        :param wait_for_cancelling: when True, always make sure the order is completely cancelled before returning.
        :param cancelling_timeout: time before raising a timeout error when waiting for an order cancel
        :return: (True, orders): True if all orders got cancelled, False if an error occurred and the list of
        cancelled orders
        """
        all_cancelled = True
        cancelled_orders = []
        for order in self.exchange_manager.exchange_personal_data.orders_manager.get_open_orders(since=since, until=until):
            if order.symbol == symbol and \
                    (side is None or order.side is side) and \
                    not (order.is_cancelled() or order.is_closed()) and \
                    (cancel_loaded_orders or order.is_from_this_octobot):
                try:
                    async with signals.remote_signal_publisher(self.exchange_manager, order.symbol, emit_trading_signals):
                        cancelled = await signals.cancel_order(
                            self.exchange_manager,
                            emit_trading_signals and signals.should_emit_trading_signal(self.exchange_manager),
                            order,
                            wait_for_cancelling=wait_for_cancelling,
                            cancelling_timeout=cancelling_timeout, )
                    if cancelled:
                        cancelled_orders.append(order)
                    all_cancelled = cancelled and all_cancelled
                except (errors.OrderCancelError, errors.UnexpectedExchangeSideOrderStateError) as err:
                    self.logger.warning(f"Skipping order cancel: {err} ({err.__class__.__name__})")
        return all_cancelled, cancelled_orders

    async def cancel_all_open_orders_with_currency(
        self, currency, emit_trading_signals=False,
        wait_for_cancelling=True,
        cancelling_timeout=octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
    ) -> bool:
        """
        Should be called only if the goal is to cancel all open orders for each traded symbol containing the
        given currency.
        :param currency: Currency to find trading pairs to cancel orders on.
        :param emit_trading_signals: when true, trading signals will be emitted
        :param wait_for_cancelling: when True, always make sure the order is completely cancelled before returning.
        :param cancelling_timeout: time before raising a timeout error when waiting for an order cancel
        :return: True if all orders got cancelled, False if an error occurred
        """
        all_cancelled = True
        symbols = util.get_pairs(self.config, currency, enabled_only=True)
        if symbols:
            for symbol in symbols:
                all_cancelled = (await self.cancel_open_orders(symbol, emit_trading_signals=emit_trading_signals,
                                                               wait_for_cancelling=wait_for_cancelling,
                                                               cancelling_timeout=cancelling_timeout))[0] \
                                and all_cancelled
        return all_cancelled

    async def cancel_all_open_orders(
            self, emit_trading_signals=False,
            wait_for_cancelling=True,
            cancelling_timeout=octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
    ) -> bool:
        """
        Cancel all open orders registered on this bot.
        :param emit_trading_signals: when true, trading signals will be emitted
        :param wait_for_cancelling: when True, always make sure the order is completely cancelled before returning.
        :param cancelling_timeout: time before raising a timeout error when waiting for an order cancel
        :return: True if all orders got cancelled, False if an error occurred
        """
        all_cancelled = True
        for order in self.exchange_manager.exchange_personal_data.orders_manager.get_open_orders():
            if not order.is_cancelled():
                try:
                    async with signals.remote_signal_publisher(
                            self.exchange_manager, order.symbol, emit_trading_signals
                    ):
                        all_cancelled = await signals.cancel_order(
                            self.exchange_manager,
                            emit_trading_signals and signals.should_emit_trading_signal(self.exchange_manager),
                            order,
                            wait_for_cancelling=wait_for_cancelling,
                            cancelling_timeout=cancelling_timeout, ) and all_cancelled
                except (errors.OrderCancelError, errors.UnexpectedExchangeSideOrderStateError) as err:
                    self.logger.warning(f"Skipping order cancel: {err} ({err.__class__.__name__})")
                    all_cancelled = False
        return all_cancelled

    async def _sell_everything(self, symbol, inverted, timeout=None):
        created_orders = []
        order_type = octobot_trading.enums.TraderOrderType.BUY_MARKET \
            if inverted else octobot_trading.enums.TraderOrderType.SELL_MARKET
        async with self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.lock:
            current_symbol_holding, current_market_quantity, _, price, symbol_market = \
                await order_util.get_pre_order_data(self.exchange_manager, symbol, timeout=timeout)
            if inverted:
                if price > 0:
                    quantity = current_market_quantity / price
                else:
                    quantity = 0
            else:
                quantity = current_symbol_holding
            for order_quantity, order_price in decimal_order_adapter.decimal_check_and_adapt_order_details_if_necessary(
                    quantity, price,
                    symbol_market):
                current_order = order_factory.create_order_instance(trader=self,
                                                                    order_type=order_type,
                                                                    symbol=symbol,
                                                                    current_price=order_price,
                                                                    quantity=order_quantity,
                                                                    price=order_price)
                created_orders.append(
                    await self.create_order(current_order))
        return created_orders

    async def sell_all(self, currencies_to_sell=None, timeout=None):
        """
        Sell every currency in portfolio for reference market using market orders.
        :param currencies_to_sell: List of currencies to sell, default values consider every currency in portfolio
        :param timeout: Timeout to get market price
        :return: The created orders
        """
        orders = []
        currency_list = self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio

        if not currencies_to_sell:
            currencies = currency_list
        else:
            currencies = [currency
                          for currency in currencies_to_sell
                          if currency in currency_list]

        for currency in currencies:
            symbol, inverted = util.get_market_pair(self.config, currency, enabled_only=True)
            if symbol:
                orders += await self._sell_everything(symbol, inverted, timeout=timeout)
        return orders

    def parse_order_id(self, order_id):
        return order_id

    def convert_order_to_trade(self, order):
        """
        Convert an order instance to Trade
        :return: the new Trade instance from order
        """
        return trade_factory.create_trade_from_order(order)

    """
    Positions
    """

    async def close_position(self, position, limit_price=None, timeout=1, emit_trading_signals=False,
                             wait_for_creation=True,
                             creation_timeout=octobot_trading.constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT):
        """
        Creates a close position order
        :param position: the position to close
        :param limit_price: the close order limit price if None uses a market order
        :param timeout: the mark price timeout
        :param emit_trading_signals: when true, trading signals will be emitted
        :param wait_for_creation: when True, always make sure the order is completely created before returning.
        On exchanges async api, a create request will return before the order is actually live on exchange, in this case
        the associated order state will make sure that the order is creating by polling the order from the exchange.
        :param creation_timeout: time before raising a timeout error when waiting for an order creation
        :return: the list of created orders
        """
        created_orders = []
        async with self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.lock:
            _, _, _, price, symbol_market = await order_util.get_pre_order_data(self.exchange_manager,
                                                                                position.symbol,
                                                                                timeout=timeout)
            for order_quantity, order_price in decimal_order_adapter.decimal_check_and_adapt_order_details_if_necessary(
                    position.get_quantity_to_close().copy_abs(), price, symbol_market):

                if limit_price is not None:
                    order_type = enums.TraderOrderType.SELL_LIMIT \
                        if position.is_long() else enums.TraderOrderType.BUY_LIMIT
                else:
                    order_type = enums.TraderOrderType.SELL_MARKET \
                        if position.is_long() else enums.TraderOrderType.BUY_MARKET

                current_order = order_factory.create_order_instance(trader=self,
                                                                    order_type=order_type,
                                                                    symbol=position.symbol,
                                                                    current_price=order_price,
                                                                    quantity=order_quantity,
                                                                    price=limit_price
                                                                    if limit_price is not None else order_price,
                                                                    reduce_only=True,
                                                                    close_position=True)
                async with signals.remote_signal_publisher(self.exchange_manager, current_order.symbol,
                                                           emit_trading_signals):
                    order = await signals.create_order(
                        self.exchange_manager,
                        emit_trading_signals and signals.should_emit_trading_signal(self.exchange_manager),
                        current_order,
                        wait_for_creation=wait_for_creation,
                        creation_timeout=creation_timeout
                    )
                created_orders.append(order)
        return created_orders

    async def withdraw(self, amount, currency):
        """
        Removes the given amount from the portfolio. Only works in simulated portfolios
        :param amount: the amount to withdraw
        :param currency: the currency to withdraw
        """
        async with self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.lock:
            await self.exchange_manager.exchange_personal_data.handle_portfolio_update_from_withdrawal(amount, currency)

    async def set_leverage(
        self, symbol: str, side: typing.Optional[enums.PositionSide], leverage: decimal.Decimal
    ) -> bool:
        """
        Updates the symbol contract leverage
        Can raise InvalidLeverageValue if leverage value is not matching requirements
        :param symbol: the symbol to update
        :param side: the side to update (TODO)
        :param leverage: the new leverage value
        :return True if leverage changed
        """
        contract = self.exchange_manager.exchange.get_pair_future_contract(symbol)
        if not contract.check_leverage_update(leverage):
            raise errors.InvalidLeverageValue(f"Trying to update leverage with {leverage} "
                                              f"but maximal value is {contract.maximum_leverage}")
        if contract.current_leverage != leverage:
            if not self.simulate:
                await self.exchange_manager.exchange.set_symbol_leverage(symbol, float(leverage))
            self.logger.info(f"Switching {symbol} leverage from {contract.current_leverage} to {leverage}")
            contract.set_current_leverage(leverage)
            return True
        return False

    async def set_symbol_take_profit_stop_loss_mode(self, symbol, new_mode: enums.TakeProfitStopLossMode):
        """
        Updates the take profit and stop loss mode for the given symbol
        Raises NotImplementedError if the endpoint is not implemented on exchange
        :param symbol: the symbol to update
        :param new_mode: the take_profit_stop_loss_mode value
        """
        contract = self.exchange_manager.exchange.get_pair_future_contract(symbol)
        if contract.take_profit_stop_loss_mode != new_mode:
            if not self.simulate:
                await self.exchange_manager.exchange.set_symbol_partial_take_profit_stop_loss(
                    symbol, contract.is_inverse_contract(), new_mode)
            self.logger.info(
                f"Switching {symbol} profit_stop_loss_mode from {contract.take_profit_stop_loss_mode} to {new_mode}"
            )
            contract.set_take_profit_stop_loss_mode(new_mode)

    async def set_margin_type(self, symbol, side, margin_type):
        """
        Updates the symbol contract margin type
        TODO: recreate position instances if any
        :param symbol: the symbol to update
        :param side: the side to update (TODO)
        :param margin_type: the new margin type (enums.MarginType)
        """
        contract = self.exchange_manager.exchange.get_pair_future_contract(symbol)
        if contract.margin_type != margin_type:
            if not self.simulate:
                await self.exchange_manager.exchange.set_symbol_margin_type(
                    symbol=symbol,
                    isolated=margin_type is enums.MarginType.ISOLATED
                )
            self.logger.info(f"Switching {symbol} margin_type from {contract.margin_type} to {margin_type}")
            contract.set_margin_type(
                is_isolated=margin_type is enums.MarginType.ISOLATED,
                is_cross=margin_type is enums.MarginType.CROSS
            )

    async def set_position_mode(self, symbol, position_mode):
        """
        Updates the symbol contract position mode
        :param symbol: the symbol to update
        :param position_mode: the new position mode (enums.PositionMode)
        """
        if self._has_open_position(symbol):
            raise errors.TooManyOpenPositionError("Can't update position mode when having open position.")
        contract = self.exchange_manager.exchange.get_pair_future_contract(symbol)
        if not self.simulate:
            await self.exchange_manager.exchange.set_symbol_position_mode(
                symbol=symbol,
                one_way=position_mode is enums.PositionMode.ONE_WAY
            )
        contract.set_position_mode(
            is_one_way=position_mode is enums.PositionMode.ONE_WAY,
            is_hedge=position_mode is enums.PositionMode.HEDGE
        )

    def _has_open_position(self, symbol):
        """
        Checks if open position exists for :symbol:
        :param symbol: the position symbol
        :return: True if open position for :symbol: exists
        """
        return len(self.exchange_manager.exchange_personal_data.positions_manager.get_symbol_positions(
            symbol=symbol)) != 0



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing
import uuid

import octobot_commons.configuration as configuration
import octobot_commons.constants as common_constants
import octobot_commons.logging as logging
import octobot_commons.timestamp_util as timestamp_util

import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.exchanges as exchanges
import octobot_trading.personal_data as personal_data
import octobot_trading.exchange_data as exchange_data
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.util as util
import octobot_trading.errors as errors
import octobot_trading.storage as storage
import trading_backend.exchanges


class ExchangeManager(util.Initializable):
    def __init__(self, config, exchange_class_string):
        super().__init__()
        self.id : str= str(uuid.uuid4())
        self.bot_id: str = None
        self.config: dict = config
        self.tentacles_setup_config = None
        self.exchange_class_string: str = exchange_class_string
        self.exchange_name: str = exchange_class_string
        self.logger = logging.get_logger(self.__class__.__name__)

        self.is_ready: bool = False
        self.is_simulated: bool = False
        self.is_backtesting: bool = False
        self.rest_only: bool = False
        self.ignore_config: bool = False
        self.use_cached_markets: bool = True
        self.market_filter: typing.Union[None, typing.Callable[[dict], bool]] = None
        self.is_spot_only: bool = False
        self.is_margin: bool = False
        self.is_future: bool = False
        self.is_sandboxed: bool = False
        self.is_trading: bool = True
        self.without_auth: bool = False
        self.check_credentials: bool = True
        self.enable_storage: bool = True
        self.proxy_config: exchanges.ProxyConfig = exchanges.ProxyConfig.default_env_var_config()

        # exchange_only is True when exchange channels are not required (therefore not created)
        self.exchange_only: bool = False

        self.backtesting = None

        self.is_trader_simulated: bool = util.is_trader_simulator_enabled(self.config)
        self.has_websocket: bool = False

        self.trader: exchanges.Trader = None
        self.exchange: exchanges.RestExchange = None
        self.exchange_backend: trading_backend.exchanges.Exchange = None
        self.is_broker_enabled: bool = False
        self.trading_modes: list = []

        self.exchange_web_socket = None

        self.client_symbols: list[str] = []
        self.client_time_frames: list[str] = []

        self.storage_manager: storage.StorageManager = storage.StorageManager(self)
        self.exchange_config: exchanges.ExchangeConfig = exchanges.ExchangeConfig(self)
        self.exchange_personal_data: personal_data.ExchangePersonalData = personal_data.ExchangePersonalData(self)
        self.exchange_symbols_data: exchange_data.ExchangeSymbolsData = exchange_data.ExchangeSymbolsData(self)

        self.debug_info = {}

    async def initialize_impl(self, exchange_config_by_exchange: typing.Optional[dict[str, dict]]):
        await exchanges.create_exchanges(self, exchange_config_by_exchange)
        if self.is_storage_enabled():
            await self.storage_manager.initialize()

    async def stop(self, warning_on_missing_elements=True, enable_logs=True):
        """
        Stops exchange manager relative tasks : websockets, trading mode, and exchange channels
        :param warning_on_missing_elements: warn on missing element
        :param enable_logs: log stopping steps
        """
        if enable_logs:
            self.logger.debug("Stopping ...")
        # stop websockets if any
        if self.has_websocket:
            if enable_logs:
                self.logger.debug("Stopping websocket ...")
            try:
                await self.exchange_web_socket.stop_sockets()
                await self.exchange_web_socket.close_sockets()
            except Exception as err:
                self.logger.exception(err, True, f"Error when stopping exchange websocket: {err}")
            if enable_logs:
                self.logger.debug("Stopped websocket")
        if self.exchange_web_socket:
            self.exchange_web_socket.clear()
            self.exchange_web_socket = None

        # stop trading modes
        if enable_logs:
            self.logger.debug("Stopping trading modes ...")
        for trading_mode in self.trading_modes:
            try:
                await trading_mode.stop()
            except Exception as err:
                self.logger.exception(err, True, f"Error when stopping trading mode: {err}")
        if enable_logs:
            self.logger.debug("Stopped trading modes")

        # stop exchange channels
        if enable_logs:
            self.logger.debug(f"Stopping exchange channels for exchange_id: {self.id} ...")
        if self.exchange is not None:
            try:
                exchange_channel.get_exchange_channels(self.id)
                await exchange_channel.stop_exchange_channels(self, should_warn=warning_on_missing_elements)
            except KeyError:
                # no exchange channel to stop
                pass
            except Exception as err:
                self.logger.exception(err, True, f"Error when stopping exchange channels: {err}")
            try:
                await self.exchange.stop()
            except Exception as err:
                self.logger.exception(err, True, f"Error when stopping exchange: {err}")
            exchanges.Exchanges.instance().del_exchange(
                self.exchange.name, self.id, should_warn=warning_on_missing_elements
            )
            self.exchange.exchange_manager = None
            self.exchange = None
        if self.exchange_personal_data is not None:
            try:
                await self.exchange_personal_data.stop()
            except Exception as err:
                self.logger.exception(err, True, f"Error when stopping exchange_personal_data: {err}")
        if self.exchange_symbols_data is not None:
            try:
                await self.exchange_symbols_data.stop()
            except Exception as err:
                self.logger.exception(err, True, f"Error when stopping exchange_symbols_data: {err}")
        if enable_logs:
            self.logger.debug(f"Stopped exchange channels for exchange_id: {self.id}")

        if enable_logs:
            self.logger.debug("Stopping storages ...")
        try:
            await self.storage_manager.stop()
        except Exception as err:
            self.logger.exception(err, True, f"Error when stopping storage_manager: {err}")

        if self.proxy_config and self.proxy_config.stop_proxy_callback is not None:
            try:
                self.proxy_config.stop_proxy_callback()
            except Exception as err:
                self.logger.exception(err, True, f"Error when stopping proxy: {err}")

        self.exchange_config = None
        self.exchange_personal_data = None
        self.exchange_symbols_data = None
        if self.exchange_backend is not None:
            try:
                self.exchange_backend.stop()
            except Exception as err:
                self.logger.exception(err, True, f"Error when stopping exchange_backend: {err}")
        if enable_logs:
            self.logger.debug("Stopping trader ...")
        if self.trader is not None:
            self.trader.clear()
            self.trader = None
        if enable_logs:
            self.logger.debug("Stopped trader")
        self.trading_modes = []
        self.backtesting = None
        if enable_logs:
            self.logger.debug("Stopped")

    async def register_trader(self, trader):
        self.trader = trader
        await self.exchange_personal_data.initialize()
        await self.exchange_config.initialize()

    def load_constants(self):
        if not self.is_backtesting:
            self._load_config_symbols_and_time_frames()
            self.exchange_config.set_config_time_frame()
            self.exchange_config.set_config_traded_pairs()
        # always call set_historical_settings
        self.exchange_config.set_historical_settings()
        if not self.exchange_only:
            self.exchange_config.initialize_exchange_event_tree()

    def need_user_stream(self):
        return self.config[common_constants.CONFIG_TRADER][common_constants.CONFIG_ENABLED_OPTION]

    def reset_exchange_symbols_data(self):
        self.exchange_symbols_data = exchange_data.ExchangeSymbolsData(self)

    def reset_exchange_personal_data(self):
        self.exchange_personal_data = personal_data.ExchangePersonalData(self)

    """
    Exchange Configuration
    """

    def check_config(self, exchange_name):
        if common_constants.CONFIG_EXCHANGE_KEY not in self.config[common_constants.CONFIG_EXCHANGES][exchange_name] \
                or common_constants.CONFIG_EXCHANGE_SECRET not in self.config[common_constants.CONFIG_EXCHANGES][exchange_name]:
            return False
        else:
            return True

    def enabled(self):
        # if we can get candlestick data
        if self.is_simulated or self.exchange.name in self.config[common_constants.CONFIG_EXCHANGES]:
            return True
        else:
            self.logger.warning(f"Exchange {self.exchange.name} is currently disabled")
            return False

    def get_exchange_symbol(self, symbol):
        return self.exchange.get_pair_from_exchange(symbol)

    def get_exchange_quote_and_base(self, symbol):
        return self.exchange.get_split_pair_from_exchange(symbol)

    def get_symbol_data(self, symbol):
        return self.exchange_symbols_data.get_exchange_symbol_data(symbol)

    def get_rest_pairs_refresh_threshold(self) -> enums.RestExchangePairsRefreshMaxThresholds:
        traded_pairs_count = len(self.exchange_config.traded_symbol_pairs)
        if traded_pairs_count < enums.RestExchangePairsRefreshMaxThresholds.FAST.value:
            return enums.RestExchangePairsRefreshMaxThresholds.FAST
        if traded_pairs_count < enums.RestExchangePairsRefreshMaxThresholds.MEDIUM.value:
            return enums.RestExchangePairsRefreshMaxThresholds.MEDIUM
        return enums.RestExchangePairsRefreshMaxThresholds.SLOW

    def _load_config_symbols_and_time_frames(self):
        if self.exchange.symbols and self.exchange.time_frames:
            self.client_symbols = list(self.exchange.symbols)
            self.client_time_frames = list(self.exchange.time_frames)
        elif not self.exchange_only:
            err_message = "Failed to load exchange symbols or time frames"
            self.logger.error(err_message)
            self._raise_exchange_load_error(err_message)

    def symbol_exists(self, symbol):
        if self.client_symbols is None:
            self.logger.error(f"Failed to load available symbols from REST exchange, impossible to check if "
                              f"{symbol} exists on {self.exchange.name}")
            return False
        return symbol in self.client_symbols

    def time_frame_exists(self, time_frame):
        if not self.client_time_frames:
            return False
        return time_frame in self.client_time_frames

    # Exceptions
    def _raise_exchange_load_error(self, desc):
        raise Exception(f"{self.exchange} - Failed to load exchange instances: {desc}")

    def get_exchange_name(self):
        return self.exchange_class_string

    def get_currently_handled_pair_with_time_frame(self):
        return len(self.exchange_config.traded_symbol_pairs) * len(self.exchange_config.available_time_frames)

    def ensure_reachability(self):
        """
        Raises UnreachableExchange if the exchange is not available
        Warning: only working in backtesting for now as self.exchange.is_unreachable
        is not updated in live mode
        """
        if self.exchange.is_unreachable:
            current_time = self.exchange.get_exchange_current_time()
            raise errors.UnreachableExchange(f"{self.exchange_name} can't be reached or is offline on the "
                                             f"{timestamp_util.convert_timestamp_to_datetime(current_time)} "
                                             f"(timestamp: {current_time})")

    def get_is_overloaded(self):
        if self.has_websocket:
            return False
        max_handled = self.exchange.get_max_handled_pair_with_time_frame()
        return max_handled != constants.INFINITE_MAX_HANDLED_PAIRS_WITH_TIMEFRAME and max_handled < \
            self.get_currently_handled_pair_with_time_frame()

    def should_decrypt_token(self):
        if configuration.has_invalid_default_config_value(
                self.config[common_constants.CONFIG_EXCHANGES][self.get_exchange_name()].get(
                    common_constants.CONFIG_EXCHANGE_KEY, ''),
                self.config[common_constants.CONFIG_EXCHANGES][self.get_exchange_name()].get(
                    common_constants.CONFIG_EXCHANGE_SECRET, '')):
            return False
        return True

    def get_exchange_credentials(self, exchange_name):
        if self.ignore_config or not self.should_decrypt_token() or self.without_auth:
            return "", "", "", "", ""
        config_exchange = self.config[common_constants.CONFIG_EXCHANGES][exchange_name]
        key = configuration.decrypt_element_if_possible(
            common_constants.CONFIG_EXCHANGE_KEY, config_exchange, None
        )
        secret = configuration.decrypt_element_if_possible(
            common_constants.CONFIG_EXCHANGE_SECRET, config_exchange, None
        )
        return (
            # remove leading and trailing ", ', newlines and whitespaces if any
            key.strip(' "').strip("'").strip("\n") if key else key,
            secret.strip(' "').strip("'").strip("\n") if secret else secret,
            configuration.decrypt_element_if_possible(
                common_constants.CONFIG_EXCHANGE_PASSWORD, config_exchange, None
            ),
            config_exchange.get(common_constants.CONFIG_EXCHANGE_UID, ""),
            config_exchange.get(common_constants.CONFIG_EXCHANGE_ACCESS_TOKEN, "")
        )

    def get_exchange_sub_account_id(self, exchange_name):
        config_exchange = self.config[common_constants.CONFIG_EXCHANGES][exchange_name]
        return config_exchange.get(common_constants.CONFIG_EXCHANGE_SUB_ACCOUNT, None)

    def is_storage_enabled(self):
        return self.enable_storage and not self.exchange_only and self.bot_id is not None

    def update_debug_info(self):
        self.debug_info = {
            "str": str(self),
            "trading_modes": str([tm.get_name() for tm in self.trading_modes]),
            "available_time_frames": str(self.exchange_config.available_time_frames),
            "watched_pairs": str(self.exchange_config.watched_pairs),
            "traded_symbol_pairs": str(self.exchange_config.traded_symbol_pairs),
        }

    def __str__(self):
        exchange_type = 'rest'
        exchange_type = 'spot only' if self.is_spot_only else exchange_type
        exchange_type = 'margin' if self.is_margin else exchange_type
        exchange_type = 'future' if self.is_future else exchange_type
        return f"[{self.__class__.__name__}] with {self.exchange.__class__.__name__ if self.exchange else '?'} " \
               f"exchange class on {self.get_exchange_name()} | {exchange_type} | " \
               f"{'authenticated | ' if self.exchange and self.exchange.authenticated() else 'unauthenticated | '}" \
               f"{'backtesting | ' if self.backtesting else ''}{'sandboxed | ' if self.is_sandboxed else ''}" \
               f"{'' if self.is_trading else 'not trading | '}" \
               f"{'websocket | ' if self.has_websocket else 'no websocket | '} id: {self.id}"



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import ccxt
import aiohttp
import aiohttp.client_reqrep

import mock
from mock import patch

import octobot_trading.exchanges.connectors as exchange_connectors
import octobot_trading.enums as enums
import octobot_trading.errors
import octobot_trading.exchange_data.contracts as contracts
import octobot_trading.exchanges.connectors.ccxt.ccxt_clients_cache as ccxt_clients_cache
import pytest

import tests.exchanges.connectors.ccxt.mock_exchanges_data as mock_exchanges_data
from tests.exchanges import exchange_manager, future_simulated_exchange_manager, set_future_exchange_fees, \
    register_market_status_mocks
from tests.exchanges.traders import future_trader, future_trader_simulator_with_default_linear, DEFAULT_FUTURE_SYMBOL, \
    DEFAULT_FUTURE_SYMBOL_MARGIN_TYPE, DEFAULT_FUTURE_SYMBOL_LEVERAGE

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def ccxt_connector(exchange_manager):
    yield exchange_connectors.CCXTConnector(exchange_manager.config, exchange_manager)


async def test_initialize_impl_with_none_symbols_and_timeframes(ccxt_connector):

    class MockCCXT:
        def __init__(self):
            self.symbols = None
            self.timeframes = None
            self.markets = {}
            self.set_markets_calls = []
            self.urls = {}

        async def load_markets(self, reload=False):
            pass

        def set_markets(self, markets):
            self.set_markets_calls.append(markets)

        def setSandboxMode(self, is_sandboxed):
            pass

    with patch.object(ccxt_connector, 'client', new=MockCCXT()) as mocked_ccxt, \
            patch.object(ccxt_connector, '_ensure_auth', new=mock.AsyncMock()) as _ensure_auth_mock:
        await ccxt_connector.initialize_impl()
        assert len(ccxt_connector.symbols) == 171   # all enabled symbols
        assert ccxt_connector.time_frames == set()
        assert mocked_ccxt.set_markets_calls in ([[]], [])  # depends on call order
        _ensure_auth_mock.assert_called_once()


async def test_initialize_impl_with_empty_symbols_and_timeframes(ccxt_connector):

    class MockCCXT:
        def __init__(self):
            self.symbols = []
            self.timeframes = []
            self.markets = {}
            self.set_markets_calls = []
            self.urls = {}

        async def load_markets(self, reload=False):
            pass

        def set_markets(self, markets):
            self.set_markets_calls.append(markets)

        def setSandboxMode(self, is_sandboxed):
            pass

    with patch.object(ccxt_connector, 'client', new=MockCCXT()) as mocked_ccxt, \
            patch.object(ccxt_connector, '_ensure_auth', new=mock.AsyncMock()) as _ensure_auth_mock:
        await ccxt_connector.initialize_impl()
        assert len(ccxt_connector.symbols) == 171   # all enabled symbols
        assert ccxt_connector.time_frames == set()
        assert mocked_ccxt.set_markets_calls in ([[]], [])  # depends on call order
        _ensure_auth_mock.assert_called_once()


async def test_initialize_impl(ccxt_connector):

    class MockCCXT:
        def __init__(self):
            self.symbols = [
                "BTC/USDT",
                "ETH/USDT",
                "ETH/BTC",
                "ETH/USDT"
            ]
            self.timeframes = [
                "1h",
                "2h",
                "4h",
                "2h"
            ]
            self.markets = {}
            self.set_markets_calls = []
            self.urls = {}

        async def load_markets(self, reload=False):
            pass

        def set_markets(self, markets):
            self.set_markets_calls.append(markets)

        def setSandboxMode(self, is_sandboxed):
            pass

    with patch.object(ccxt_connector, 'client', new=MockCCXT()) as mocked_ccxt, \
        patch.object(ccxt_connector, '_ensure_auth', new=mock.AsyncMock()) as _ensure_auth_mock:
        ccxt_connector.exchange_manager.exchange.INCLUDE_DISABLED_SYMBOLS_IN_AVAILABLE_SYMBOLS = True
        await ccxt_connector.initialize_impl()
        assert len(ccxt_connector.symbols) == 541   # all enabled + diasabled symbols
        assert ccxt_connector.time_frames == {
            "1h",
            "2h",
            "4h",
        }
        assert mocked_ccxt.set_markets_calls in ([[]], [])  # depends on call order
        _ensure_auth_mock.assert_called_once()


async def test_set_symbol_partial_take_profit_stop_loss(ccxt_connector):
    with pytest.raises(NotImplementedError):
        await ccxt_connector.set_symbol_partial_take_profit_stop_loss("BTC/USDT", False,
                                                                     enums.TakeProfitStopLossMode.PARTIAL)


async def test_get_ccxt_order_type(ccxt_connector):
    with pytest.raises(RuntimeError):
        ccxt_connector.get_ccxt_order_type(None)
    with pytest.raises(RuntimeError):
        ccxt_connector.get_ccxt_order_type(enums.TraderOrderType.UNKNOWN)
    assert ccxt_connector.get_ccxt_order_type(enums.TraderOrderType.BUY_LIMIT) == enums.TradeOrderType.LIMIT.value
    assert ccxt_connector.get_ccxt_order_type(enums.TraderOrderType.STOP_LOSS_LIMIT) == enums.TradeOrderType.LIMIT.value
    assert ccxt_connector.get_ccxt_order_type(enums.TraderOrderType.TRAILING_STOP) == enums.TradeOrderType.MARKET.value
    assert ccxt_connector.get_ccxt_order_type(enums.TraderOrderType.SELL_MARKET) == enums.TradeOrderType.MARKET.value


async def test_get_trade_fee(exchange_manager, future_trader_simulator_with_default_linear):
    spot_fees_value = 0.001
    future_symbol = "BTC/USDT:USDT"
    future_fees_value = 0.0004
    spot_symbol = "BTC/USDT"
    config, fut_exchange_manager_inst, trader_inst, default_contract = future_trader_simulator_with_default_linear
    fut_exchange_manager_inst.is_future = True
    fut_ccxt_exchange = exchange_connectors.CCXTConnector(config, fut_exchange_manager_inst)
    spot_ccxt_exchange = exchange_connectors.CCXTConnector(exchange_manager.config, exchange_manager)

    # spot trading
    spot_ccxt_exchange.client.options['defaultType'] = enums.ExchangeTypes.SPOT.value
    await spot_ccxt_exchange.client.load_markets()
    assert spot_fees_value / 5 <= spot_ccxt_exchange.client.markets[spot_symbol]['taker'] <= spot_fees_value * 5
    assert spot_ccxt_exchange.get_trade_fee(spot_symbol, enums.TraderOrderType.BUY_LIMIT, decimal.Decimal("0.45"),
                                            decimal.Decimal(10000), "taker") == \
           _get_fees("taker", "BTC", 0.001, decimal.Decimal("0.00045"))
    assert spot_ccxt_exchange.get_trade_fee(spot_symbol, enums.TraderOrderType.SELL_LIMIT, decimal.Decimal("0.45"),
                                            decimal.Decimal(10000), "maker") == \
           _get_fees("maker", "USDT", 0.001, decimal.Decimal("4.5"))

    # future trading
    fut_ccxt_exchange.client.options['defaultType'] = enums.ExchangeTypes.FUTURE.value

    if forced_markets := mock_exchanges_data.MOCKED_EXCHANGE_SYMBOL_DETAILS.get(
        fut_exchange_manager_inst.exchange_name, None
    ):
        register_market_status_mocks(fut_exchange_manager_inst.exchange_name)
    await fut_ccxt_exchange.load_symbol_markets()
    # enforce taker and maker values
    set_future_exchange_fees(fut_ccxt_exchange, future_symbol, taker=future_fees_value, maker=future_fees_value)
    assert future_fees_value / 5 <= fut_ccxt_exchange.client.markets[future_symbol]['taker'] <= future_fees_value * 5
    # linear
    assert fut_ccxt_exchange.get_trade_fee(future_symbol, enums.TraderOrderType.BUY_LIMIT, decimal.Decimal("0.45"),
                                           decimal.Decimal(10000), "taker") == \
           _get_fees("taker", "USDT", future_fees_value, decimal.Decimal("1.800000"))
    # inverse
    fut_ccxt_exchange.client.markets[future_symbol]["inverse"] = True
    fut_ccxt_exchange.client.markets[future_symbol]["linear"] = False
    contract = contracts.FutureContract(pair=future_symbol,
                                        margin_type=enums.MarginType.ISOLATED,
                                        contract_type=enums.FutureContractType.INVERSE_PERPETUAL)
    fut_exchange_manager_inst.exchange.pair_contracts[future_symbol] = contract
    assert fut_ccxt_exchange.get_trade_fee(future_symbol, enums.TraderOrderType.BUY_LIMIT, decimal.Decimal("0.45"),
                                           decimal.Decimal(10000), "taker") == \
           _get_fees("taker", "BTC", future_fees_value, decimal.Decimal("0.00018"))


async def test_error_describer(ccxt_connector):
    with pytest.raises(ZeroDivisionError, match="plop"):
        # random error is just forwarded
        with ccxt_connector.error_describer():
            raise ZeroDivisionError("plop")
    with pytest.raises(ccxt.DDoSProtection):
        # forwarded ccxt error
        with ccxt_connector.error_describer():
            raise ccxt.DDoSProtection("plop")
    with pytest.raises(octobot_trading.errors.FailedRequest):
        # transformed ccxt error
        with ccxt_connector.error_describer():
            raise ccxt.InvalidNonce("plop")
    with pytest.raises(octobot_trading.errors.FailedRequest):
        # transformed ccxt error
        with ccxt_connector.error_describer():
            raise ccxt.RequestTimeout("plop")
    with pytest.raises(ccxt.ExchangeError):
        # forwarded ccxt error
        with ccxt_connector.error_describer():
            raise ccxt.ExchangeError("plop")
    with mock.patch.object(
        ccxt_connector.exchange_manager.exchange, "is_authentication_error", mock.Mock(return_value=True)
    ) as is_authentication_error_mock:
        with pytest.raises(octobot_trading.errors.AuthenticationError):
            # transformed ccxt error
            with ccxt_connector.error_describer():
                raise ccxt.ExchangeError("plop")
        is_authentication_error_mock.assert_called_once()

    # proxied errors
    with pytest.raises(octobot_trading.errors.AuthenticationError):
        # transformed ccxt error
        with ccxt_connector.error_describer():
            raise ccxt.AuthenticationError("plop")
    with pytest.raises(octobot_trading.errors.FailedRequest):
        # transformed ccxt error
        with ccxt_connector.error_describer():
            raise ccxt.ExchangeNotAvailable("plop")
    # default host, not using proxy exception for generic exception
    with pytest.raises(octobot_trading.errors.AuthenticationError):
        # proxy connection error
        with ccxt_connector.error_describer():
            raise ccxt.AuthenticationError from aiohttp.ClientConnectionError(
                aiohttp.client_reqrep.ConnectionKey("host", 11, True, True, None, None, None), OSError("plop")
            )
    # default host, using proxy exception for proxy exception
    with pytest.raises(octobot_trading.errors.ExchangeProxyError):
        # proxy connection error
        with ccxt_connector.error_describer():
            raise ccxt.AuthenticationError from aiohttp.ClientProxyConnectionError(
                aiohttp.client_reqrep.ConnectionKey("host", 11, True, True, None, None, None), OSError("plop")
            )
    ccxt_connector.exchange_manager.proxy_config.proxy_host = "host"
    with pytest.raises(octobot_trading.errors.ExchangeProxyError):
        # proxy connection error
        with ccxt_connector.error_describer():
            raise ccxt.AuthenticationError from aiohttp.ClientProxyConnectionError(
                aiohttp.client_reqrep.ConnectionKey("host", 11, True, True, None, None, None), OSError("plop")
            )
    with mock.patch.object(
        ccxt_connector.exchange_manager.proxy_config, "get_last_proxied_request_url",
            mock.Mock(return_value="https///plop.com/coucou")
    ) as get_last_proxied_request_url_mock:
        with pytest.raises(octobot_trading.errors.AuthenticationError, match="plop"):
            # not proxied request error
            with ccxt_connector.error_describer():
                raise ccxt.AuthenticationError("plop")
        get_last_proxied_request_url_mock.assert_called_once()
        get_last_proxied_request_url_mock.reset_mock()
        with pytest.raises(octobot_trading.errors.AuthenticationError, match="plop"):
            # not proxied request error
            ccxt_connector.client.last_request_url = "not_plop"
            with ccxt_connector.error_describer():
                raise ccxt.AuthenticationError("plop")
        get_last_proxied_request_url_mock.assert_called_once()
        get_last_proxied_request_url_mock.reset_mock()
        with pytest.raises(octobot_trading.errors.AuthenticationError, match="\[Proxied\] plop"):
            # proxied request error: add prefix
            ccxt_connector.client.last_request_url = "https///plop.com/coucou"
            with ccxt_connector.error_describer():
                raise ccxt.AuthenticationError("plop")
        get_last_proxied_request_url_mock.assert_called_once()
        get_last_proxied_request_url_mock.reset_mock()



def _get_fees(type, currency, rate, cost):
    return {
        enums.FeePropertyColumns.TYPE.value: type,
        enums.FeePropertyColumns.CURRENCY.value: currency,
        enums.FeePropertyColumns.RATE.value: rate,
        enums.FeePropertyColumns.COST.value: decimal.Decimal(str(cost)),
        enums.FeePropertyColumns.IS_FROM_EXCHANGE.value: False,
    }



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.exchange_data import funding
from octobot_trading.exchange_data.funding import (
    FundingUpdaterSimulator,
    FundingUpdater,
    FundingManager,
    FundingProducer,
    FundingChannel,
)
from octobot_trading.exchange_data import kline
from octobot_trading.exchange_data.kline import (
    KlineUpdaterSimulator,
    KlineProducer,
    KlineChannel,
    KlineManager,
    KlineUpdater,
)
from octobot_trading.exchange_data import ohlcv
from octobot_trading.exchange_data.ohlcv import (
    CandlesManager,
    PreloadedCandlesManager,
    get_symbol_close_candles,
    get_symbol_open_candles,
    get_symbol_high_candles,
    get_symbol_low_candles,
    get_symbol_volume_candles,
    get_symbol_time_candles,
    get_candle_as_list,
    OHLCVUpdaterSimulator,
    OHLCVProducer,
    OHLCVChannel,
    OHLCVUpdater,
)
from octobot_trading.exchange_data import order_book
from octobot_trading.exchange_data.order_book import (
    OrderBookUpdater,
    OrderBookProducer,
    OrderBookChannel,
    OrderBookTickerProducer,
    OrderBookTickerChannel,
    OrderBookManager,
    OrderBookUpdaterSimulator,
)
from octobot_trading.exchange_data import prices
from octobot_trading.exchange_data.prices import (
    MarkPriceUpdaterSimulator,
    MarkPriceProducer,
    MarkPriceChannel,
    PricesManager,
    calculate_mark_price_from_recent_trade_prices,
    MarkPriceUpdater,
    PriceEventsManager,
)
from octobot_trading.exchange_data import recent_trades
from octobot_trading.exchange_data.recent_trades import (
    RecentTradeProducer,
    RecentTradeChannel,
    LiquidationsProducer,
    LiquidationsChannel,
    RecentTradesManager,
    RecentTradeUpdater,
    RecentTradeUpdaterSimulator,
)
from octobot_trading.exchange_data import ticker
from octobot_trading.exchange_data.ticker import (
    TickerManager,
    TickerUpdater,
    TickerProducer,
    TickerChannel,
    MiniTickerProducer,
    MiniTickerChannel,
    TickerUpdaterSimulator,
)
from octobot_trading.exchange_data import contracts
from octobot_trading.exchange_data.contracts import (
    MarginContract,
    FutureContract,
    update_contracts_from_positions,
    update_future_contract_from_dict,
    create_default_future_contract,
)
from octobot_trading.exchange_data import exchange_symbol_data
from octobot_trading.exchange_data.exchange_symbol_data import (
    ExchangeSymbolData,
)
from octobot_trading.exchange_data import exchange_symbols_data
from octobot_trading.exchange_data.exchange_symbols_data import (
    ExchangeSymbolsData,
)

import octobot_trading.constants as trading_constants
import octobot_backtesting.enums as backtesting_enums

UNAUTHENTICATED_UPDATER_PRODUCERS = [OHLCVUpdater, OrderBookUpdater, RecentTradeUpdater, TickerUpdater,
                                     KlineUpdater, MarkPriceUpdater, FundingUpdater]
UNAUTHENTICATED_UPDATER_SIMULATOR_PRODUCERS = {
        trading_constants.OHLCV_CHANNEL: OHLCVUpdaterSimulator,
        trading_constants.ORDER_BOOK_CHANNEL: OrderBookUpdaterSimulator,
        trading_constants.RECENT_TRADES_CHANNEL: RecentTradeUpdaterSimulator,
        trading_constants.TICKER_CHANNEL: TickerUpdaterSimulator,
        trading_constants.KLINE_CHANNEL: KlineUpdaterSimulator,
        trading_constants.MARK_PRICE_CHANNEL: MarkPriceUpdaterSimulator,
        trading_constants.FUNDING_CHANNEL: FundingUpdaterSimulator
    }

# Required data to run updater (requires at least one per list)
SIMULATOR_PRODUCERS_TO_POSSIBLE_DATA_TYPE = {
    trading_constants.OHLCV_CHANNEL: [backtesting_enums.ExchangeDataTables.OHLCV],
    trading_constants.ORDER_BOOK_CHANNEL: [backtesting_enums.ExchangeDataTables.ORDER_BOOK],
    trading_constants.RECENT_TRADES_CHANNEL: [backtesting_enums.ExchangeDataTables.RECENT_TRADES,
                                              backtesting_enums.ExchangeDataTables.OHLCV],
    trading_constants.TICKER_CHANNEL: [backtesting_enums.ExchangeDataTables.TICKER,
                                       backtesting_enums.ExchangeDataTables.OHLCV],
    trading_constants.KLINE_CHANNEL: [backtesting_enums.ExchangeDataTables.KLINE],
    # trading_constants.FUNDING_CHANNEL: [backtesting_enums.ExchangeDataTables.FUNDING] only hard coded value for now
}

# Required data to run real data updater (requires each per list)
SIMULATOR_PRODUCERS_TO_REAL_DATA_TYPE = {
    trading_constants.OHLCV_CHANNEL: [backtesting_enums.ExchangeDataTables.OHLCV],
    trading_constants.ORDER_BOOK_CHANNEL: [backtesting_enums.ExchangeDataTables.ORDER_BOOK],
    trading_constants.RECENT_TRADES_CHANNEL: [backtesting_enums.ExchangeDataTables.RECENT_TRADES],
    trading_constants.TICKER_CHANNEL: [backtesting_enums.ExchangeDataTables.TICKER],
    trading_constants.KLINE_CHANNEL: [backtesting_enums.ExchangeDataTables.KLINE],
    trading_constants.FUNDING_CHANNEL: [backtesting_enums.ExchangeDataTables.FUNDING],
}

__all__ = [
    "FundingUpdaterSimulator",
    "FundingUpdater",
    "FundingManager",
    "FundingProducer",
    "FundingChannel",
    "KlineUpdaterSimulator",
    "KlineProducer",
    "KlineChannel",
    "KlineManager",
    "KlineUpdater",
    "CandlesManager",
    "PreloadedCandlesManager",
    "get_symbol_close_candles",
    "get_symbol_open_candles",
    "get_symbol_high_candles",
    "get_symbol_low_candles",
    "get_symbol_volume_candles",
    "get_symbol_time_candles",
    "get_candle_as_list",
    "OHLCVUpdaterSimulator",
    "OHLCVProducer",
    "OHLCVChannel",
    "OHLCVUpdater",
    "OrderBookUpdater",
    "OrderBookProducer",
    "OrderBookChannel",
    "OrderBookTickerProducer",
    "OrderBookTickerChannel",
    "OrderBookManager",
    "OrderBookUpdaterSimulator",
    "MarkPriceUpdaterSimulator",
    "MarkPriceProducer",
    "MarkPriceChannel",
    "PricesManager",
    "calculate_mark_price_from_recent_trade_prices",
    "MarkPriceUpdater",
    "PriceEventsManager",
    "RecentTradeProducer",
    "RecentTradeChannel",
    "LiquidationsProducer",
    "LiquidationsChannel",
    "RecentTradesManager",
    "RecentTradeUpdater",
    "RecentTradeUpdaterSimulator",
    "TickerManager",
    "TickerUpdater",
    "TickerProducer",
    "TickerChannel",
    "MiniTickerProducer",
    "MiniTickerChannel",
    "TickerUpdaterSimulator",
    "MarginContract",
    "FutureContract",
    "update_contracts_from_positions",
    "update_future_contract_from_dict",
    "create_default_future_contract",
    "ExchangeSymbolsData",
    "ExchangeSymbolData",
    "UNAUTHENTICATED_UPDATER_PRODUCERS",
    "UNAUTHENTICATED_UPDATER_SIMULATOR_PRODUCERS",
    "SIMULATOR_PRODUCERS_TO_POSSIBLE_DATA_TYPE",
    "SIMULATOR_PRODUCERS_TO_REAL_DATA_TYPE",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.tree as commons_tree


async def wait_for_topic_init(exchange_manager, timeout, topic, symbol=None, time_frame=None, symbols=None):
    if symbols:
        return await asyncio.gather(
            *[
                wait_for_topic_init(exchange_manager, timeout, topic, symbol=local_symbol, time_frame=None)
                for local_symbol in symbols
            ]
        )
    else:
        return await commons_tree.EventProvider.instance().wait_for_event(
            exchange_manager.bot_id,
            commons_tree.get_exchange_path(
                exchange_manager.exchange_name,
                topic,
                symbol=symbol,
                time_frame=time_frame
            ),
            timeout
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_commons.logging as logging
import octobot_commons.databases as databases
import octobot_commons.constants as common_constants

import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_trading.errors as errors
import octobot_trading.constants as constants
import octobot_trading.storage as storage
import octobot_trading.enums as trading_enums
import octobot_trading.modes.script_keywords.basic_keywords as basic_keywords
import octobot_trading.personal_data as trading_personal_data
import octobot_trading.exchanges.util.exchange_util as exchange_util


def get_required_candles_count(trading_mode_class, tentacles_setup_config):
    return tentacles_manager_api.get_tentacle_config(tentacles_setup_config, trading_mode_class).get(
        constants.CONFIG_CANDLES_HISTORY_SIZE_KEY,
        common_constants.DEFAULT_IGNORED_VALUE
    )


async def clear_simulated_orders_cache(trading_mode):
    await basic_keywords.clear_orders_cache(
        databases.RunDatabasesProvider.instance().get_orders_db(
            trading_mode.bot_id,
            storage.get_account_type_suffix_from_exchange_manager(trading_mode.exchange_manager),
            trading_mode.exchange_manager.exchange_name
        )
    )


async def clear_plotting_cache(trading_mode):
    await basic_keywords.clear_symbol_plot_cache(
        databases.RunDatabasesProvider.instance().get_symbol_db(
            trading_mode.bot_id,
            trading_mode.exchange_manager.exchange_name, trading_mode.symbol
        )
    )


def get_assets_requiring_extra_price_data_to_convert(exchange_manager, sellable_assets: list, target_asset: str) -> set:
    missing_price_assets = set()
    for asset in sellable_assets:
        portfolio = exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio
        if asset != target_asset and asset in portfolio and portfolio[asset].available:
            try:
                if exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
                   value_converter.evaluate_value(
                       asset, constants.ONE, raise_error=True, target_currency=target_asset, init_price_fetchers=False
                   ) == constants.ZERO:
                    # 0 is default value in backtesting
                    missing_price_assets.add(asset)
            except errors.MissingPriceDataError:
                # converter is not enough
                missing_price_assets.add(asset)
    return missing_price_assets


async def convert_assets_to_target_asset(trading_mode, sellable_assets: list, target_asset: str, tickers: dict) -> list:
    created_orders = []
    for asset in sorted(sellable_assets):
        new_orders = await convert_asset_to_target_asset(
            trading_mode, asset, target_asset, tickers, asset_amount=None
        )
        created_orders += new_orders
    return created_orders


async def convert_asset_to_target_asset(
    trading_mode, asset: str, target_asset: str, tickers: dict, asset_amount=None
) -> list:
    if asset == target_asset:
        return []
    created_orders = []
    tickers = tickers or {}
    portfolio = trading_mode.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio
    if asset in portfolio and portfolio[asset].available:
        created_orders.extend(
            await convert_with_market_or_limit_order(trading_mode, asset, target_asset, tickers, asset_amount)
        )
    return created_orders


async def convert_with_market_or_limit_order(
    trading_mode, asset: str, target_asset: str, tickers: dict, asset_amount=None
) -> list:
    # get symbol of the order
    symbol, order_type = _get_associated_symbol_and_order_type(trading_mode, asset, target_asset)
    if symbol is None:
        # can't convert asset into target_asset
        trading_mode.logger.warning(
            f"Impossible to convert {asset} into {target_asset}: no associated trading pair "
            f"on {trading_mode.exchange_manager.exchange_name}"
        )
        return []

    # get symbol price
    price_base = asset
    price_target = target_asset
    if order_type is trading_enums.TraderOrderType.BUY_MARKET:
        price_base = target_asset
        price_target = asset

    price = trading_personal_data.get_asset_price_from_converter_or_tickers(
        trading_mode.exchange_manager, price_base, price_target, symbol, tickers
    )

    if not price:
        # can't get price, should not happen as symbol is in client_symbols
        trading_mode.logger.error(
            f"Impossible to convert {asset} into {target_asset}: {symbol} ticker can't be fetched"
        )
        return []

    if trading_personal_data.get_trade_order_type(order_type) is not trading_enums.TradeOrderType.MARKET:
        # can't use market orders: use limit orders with price a bit under the current price to instant fill it.
        price = get_instantly_filled_limit_order_adapted_price(price, order_type)

    # get order quantity
    quantity = _get_available_or_target_quantity(trading_mode, symbol, order_type, price, asset_amount)
    symbol_market = trading_mode.exchange_manager.exchange.get_market_status(symbol, with_fixer=False)
    created_orders = []
    for order_quantity, order_price in \
            trading_personal_data.decimal_check_and_adapt_order_details_if_necessary(
                quantity,
                price,
                symbol_market
            ):
        # create order
        order = trading_personal_data.create_order_instance(
            trader=trading_mode.exchange_manager.trader,
            order_type=order_type,
            symbol=symbol,
            current_price=price,
            quantity=order_quantity,
            price=order_price
        )
        initialized_order = await trading_mode.create_order(order)
        if isinstance(initialized_order, trading_personal_data.LimitOrder) and initialized_order.simulated:
            # on simulator, this order should be instantly filled now as its price is meant to be instantly filled
            await initialized_order.on_fill()
        created_orders.append(initialized_order)
    return created_orders


def get_instantly_filled_limit_order_adapted_price(
    price: decimal.Decimal, order_type: trading_enums.TraderOrderType
) -> decimal.Decimal:
    price_delta = price * constants.INSTANT_FILLED_LIMIT_ORDER_PRICE_DELTA
    if order_type is trading_enums.TraderOrderType.SELL_LIMIT:
        price -= price_delta
    elif order_type is trading_enums.TraderOrderType.BUY_LIMIT:
        price += price_delta
    else:
        logging.get_logger(__name__).error(
            f"Unhandled order type in convertor limit order price adapter: {order_type}"
        )
    return price


def get_instantly_filled_limit_order_adapted_price_and_quantity(
    price: decimal.Decimal, quantity: decimal.Decimal, order_type: trading_enums.TraderOrderType
) -> (decimal.Decimal, decimal.Decimal):
    adapted_price = get_instantly_filled_limit_order_adapted_price(price, order_type)
    origin_cost = price * quantity
    # keep the same total cost, adapt quantity
    adapted_quantity = origin_cost / adapted_price
    return adapted_price, adapted_quantity


def _get_associated_symbol_and_order_type(trading_mode, asset: str, target_asset: str) \
     -> (str, trading_enums.TraderOrderType):
    symbol, reversed_symbol = exchange_util.get_associated_symbol(trading_mode.exchange_manager, asset, target_asset)
    if symbol is None:
        return None, None
    order_type = trading_enums.TraderOrderType.BUY_MARKET if reversed_symbol else \
        trading_enums.TraderOrderType.SELL_MARKET
    if not trading_mode.exchange_manager.exchange.is_market_open_for_order_type(symbol, order_type):
        # can't use market orders: use limit orders instead
        order_type = trading_enums.TraderOrderType.BUY_LIMIT if order_type is trading_enums.TraderOrderType.BUY_MARKET \
            else trading_enums.TraderOrderType.SELL_LIMIT
        if not trading_mode.exchange_manager.exchange.is_market_open_for_order_type(symbol, order_type):
            # can't convert asset: still try with limit orders but it will probably fail
            trading_mode.logger.error(
                f"Both market and {order_type} order are currently unsupported. Trying limit orders anyway."
            )
    return symbol, order_type


def _get_available_or_target_quantity(trading_mode, symbol, order_type, price, asset_amount) -> decimal.Decimal:
    side = (
        trading_enums.TradeOrderSide.SELL
        if order_type in (trading_enums.TraderOrderType.SELL_MARKET, trading_enums.TraderOrderType.SELL_LIMIT)
        else trading_enums.TradeOrderSide.BUY
    )

    currency_available, market_available, market_quantity = trading_personal_data.get_portfolio_amounts(
        trading_mode.exchange_manager, symbol, price, portfolio_type=common_constants.PORTFOLIO_AVAILABLE
    )
    if asset_amount is None:
        quantity = currency_available if side is trading_enums.TradeOrderSide.SELL else market_quantity
    else:
        try:
            quantity = asset_amount if side is trading_enums.TradeOrderSide.SELL else (asset_amount / price)
        except (decimal.DivisionByZero, decimal.InvalidOperation):
            quantity = constants.ZERO

    adapted_quantity = trading_personal_data.decimal_adapt_order_quantity_because_fees(
        trading_mode.exchange_manager, symbol, order_type, quantity, price,
        trading_enums.ExchangeConstantsMarketPropertyColumns.TAKER, side,
        currency_available if side is trading_enums.TradeOrderSide.SELL else market_available
    )
    return adapted_quantity


async def notify_portfolio_optimization_complete():
    try:
        import octobot_services.api as services_api
        import octobot_services.enums as services_enum
        title = "Portfolio optimization complete"
        alert_content = "Your portfolio funds are now in a optimal configuration to start your strategy"
        await services_api.send_notification(services_api.create_notification(
            alert_content,
            title=title,
            markdown_text=alert_content,
            category=services_enum.NotificationCategory.TRADES)
        )
    except ImportError as e:
        logging.get_logger(__name__).exception(e, True, f"Impossible to send notification: {e}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.asset as asset_class


class FutureAsset(asset_class.Asset):
    def __init__(self, name, available, total,
                 initial_margin=constants.ZERO,
                 wallet_balance=constants.ZERO,
                 position_margin=constants.ZERO,
                 order_margin=constants.ZERO,
                 unrealized_pnl=constants.ZERO):
        """
        Future Asset
        :param name: the asset name
        :param available: the margin available (wallet balance - used margin)
        :param total: the margin balance (wallet balance + unrealized PNL)
        :param initial_margin: the total initial margin required with current mark price
        :param wallet_balance: the wallet balance
        :param position_margin: margin required for positions with current mark price
        :param order_margin: margin required for open orders with current mark price
        :param unrealized_pnl: the unrealized profit and loss
        """
        super().__init__(name, available, total)
        # Unrealized profit and loss
        self.unrealized_pnl = unrealized_pnl

        # total initial margin required with current mark price
        self.initial_margin = initial_margin

        # The Wallet balance. When using Cross Margin, the number minus the unclosed loss is the real wallet balance.
        self.wallet_balance = wallet_balance if wallet_balance != constants.ZERO else total

        # margin required for positions with current mark price
        self.position_margin = position_margin

        # margin required for open orders with current mark price
        self.order_margin = order_margin

    def __str__(self):
        return super().__str__() + " | " \
                                   f"Initial Margin: {float(self.initial_margin)} | " \
                                   f"Wallet Balance: {float(self.wallet_balance)} | " \
                                   f"Unrealized PNL: {float(self.unrealized_pnl)} | " \
                                   f"Order Margin: {float(self.order_margin)} | " \
                                   f"Position Margin: {float(self.position_margin)}"

    def __eq__(self, other):
        if isinstance(other, FutureAsset):
            return self.available == other.available and self.total == other.total and \
                   self.initial_margin == other.initial_margin and self.wallet_balance == other.wallet_balance and \
                   self.position_margin == other.position_margin and self.order_margin == other.order_margin
        return False

    def _specific_restore_unavailable_from_other(self, other_asset):
        if other_asset.initial_margin != constants.ZERO:
            self.initial_margin = self.initial_margin + other_asset.initial_margin
        # update wallet_balance when handling cross positions
        if other_asset.position_margin != constants.ZERO:
            self.position_margin = self.position_margin + other_asset.position_margin
        if other_asset.order_margin != constants.ZERO:
            self.order_margin = self.order_margin + other_asset.order_margin


    def restore(self, old_asset):
        """
        Restore asset from previous state
        :param old_asset: previous asset state
        """
        super().restore(old_asset)
        self.unrealized_pnl = old_asset.unrealized_pnl
        self.initial_margin = old_asset.initial_margin
        self.wallet_balance = old_asset.wallet_balance
        self.position_margin = old_asset.position_margin
        self.order_margin = old_asset.order_margin

    def update(self, total=constants.ZERO, available=constants.ZERO, position_margin=constants.ZERO,
               unrealized_pnl=constants.ZERO, initial_margin=constants.ZERO):
        """
        Update asset portfolio
        :param total: the wallet balance delta
        :param initial_margin: the initial margin delta
        :param position_margin: the position margin delta
        :param available: the order margin delta
        :param unrealized_pnl: the unrealized pnl delta
        :return: True if updated
        """
        if available == constants.ZERO and position_margin == constants.ZERO \
                and total == constants.ZERO and unrealized_pnl == constants.ZERO:
            return False

        with self.update_or_restore():
            self.initial_margin += self._ensure_update_validity(self.initial_margin, initial_margin)
            self.position_margin = self._ensure_not_negative(self.position_margin + position_margin)
            self.order_margin = self._ensure_not_negative(self.order_margin + available)
            self.unrealized_pnl += unrealized_pnl
            self.wallet_balance += self._ensure_update_validity(self.wallet_balance, total)
            self._update_available()
            self._update_total()
            return True

    def set(self, total=constants.ZERO, available=None, initial_margin=constants.ZERO, position_margin=constants.ZERO,
            unrealized_pnl=constants.ZERO, order_margin=constants.ZERO, margin_balance=None):
        """
        Set available, total, initial_margin, wallet_balance, position_margin and maintenance_margin
        values for portfolio asset
        :param total: the wallet balance value
        :param available: the available margin balance value
        :param initial_margin: the initial margin value
        :param position_margin: the position margin value
        :param unrealized_pnl: the unrealized pnl value
        :param order_margin: the order margin value
        :param margin_balance: the margin balance value
        :return: True if updated
        """
        if position_margin == self.position_margin and order_margin == self.order_margin \
                and total == self.wallet_balance:
            return False

        with self.update_or_restore():
            self.initial_margin = initial_margin
            self.position_margin = position_margin
            self.order_margin = order_margin
            self.unrealized_pnl = unrealized_pnl
            self.wallet_balance = total
            if available is not None:
                self.available = available
            else:
                self._update_available()
            if margin_balance is not None:
                self.total = margin_balance
            else:
                self._update_total()
            return True

    def set_unrealized_pnl(self, unrealized_pnl):
        """
        Sets the unrealized pnl value and updates the total (margin balance) value
        :param unrealized_pnl: the new unrealized pnl value
        """
        self.unrealized_pnl = unrealized_pnl
        self._update_total()

    def update_realised_pnl(self, realized_pnl_update):
        """
        Updates the realized pnl value
        :param realized_pnl_update: the realized pnl update
        """
        self.wallet_balance += self._ensure_update_validity(self.wallet_balance, realized_pnl_update)
        self._update_total()
        self._update_available()

    def reset(self):
        """
        Reset asset portfolio to zero
        """
        self.set(total=constants.ZERO, available=constants.ZERO,
                 initial_margin=constants.ZERO, margin_balance=constants.ZERO,
                 unrealized_pnl=constants.ZERO, order_margin=constants.ZERO, position_margin=constants.ZERO)

    def _update_total(self):
        """
        Update total (margin balance) value with wallet balance + unrealized pnl
        """
        self.total = self._ensure_update_validity(constants.ZERO, self.wallet_balance + self.unrealized_pnl)

    def _update_available(self):
        """
        Update available (margin available) value with wallet balance - used margin
        """
        self.available = self._ensure_update_validity(constants.ZERO,
                                                      self.wallet_balance - self.position_margin - self.order_margin)
